# coding: utf-8
"""
    BasePlane.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""
import numpy as np
import statsmodels.api as sm
from LmfitThreadSafe import minimize, Parameters
from ThreeDimUtils import compute_plane
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
import molass_legacy.KekLib.DebugPlot as plt
# import matplotlib.pyplot as plt

LOW_PERCENTILE_BASE_1ST = 30
LOW_PERCENTILE_BASE_FIN = 15
def debug_plot(title, data, x, ix, iy, e_curve, smp_i, A, B, C):
    from mpl_toolkits.mplot3d import Axes3D
    xs = x[ix]
    ys = np.arange( iy.start, iy.stop )
    smp_x = x[smp_i]

    fig = plt.figure( figsize=( 10, 10 ) )
    ax = fig.add_subplot(111, projection='3d')
    ax.set_title(title)
    # ax.set_xlim(x[0], x[-1])

    xx, yy = np.meshgrid( np.linspace(xs[0], xs[-1], 10), np.linspace(ys[0], ys[-1], 10) )
    zz = xx * A + yy * B + C
    ax.plot_surface(xx, yy, zz, color='red', alpha=0.1 )

    y_ = ys
    x_ = np.ones(len(y_)) * smp_x
    z_ = e_curve.y
    ax.plot(x_, y_, z_, color='blue')
    for b in e_curve.boundaries:
        xp = smp_x
        zp = data[smp_i,b]
        # ax.plot([xp], [b], [zp], marker='o', color='yellow')
        # workaround for matplotlib 3 bug?
        ax.plot([xp, xp], [b, b], [zp, zp], marker='o', color='yellow')

    x_ = xs

    for k, info in enumerate( e_curve.peak_info ):
        peak_i  = int( info[1]+0.5 )
        y_ = np.ones(len(x_)) * peak_i
        z_ = data[ix,peak_i]
        ax.plot(x_, y_, z_, ':', color='green')
        xp = smp_x
        yp = peak_i
        zp = data[smp_i,peak_i]
        # ax.plot([xp], [yp], [zp], marker='o', color='red')
        # workaround for matplotlib 3 bug?
        ax.plot([xp, xp], [yp, yp], [zp, zp], marker='o', color='red')

    fig.tight_layout()
    plt.show()

class LowPercentilePlane:
    def __init__(self, data, x, ecurve_y, ix, iy, smp_i, debug=False):
        """
            z = A * x + B * y + C
            
        """
        usable_data_orig = np.array( data[ ix, iy ] )
        usable_data = np.array( usable_data_orig )

        for i in range(3):
            zpp = np.percentile( usable_data, LOW_PERCENTILE_BASE_1ST )
            # print( 'usable_data < zpp=', usable_data < zpp )
            wpp = np.where( usable_data < zpp )
            # print( 'wpp=', wpp )

            x_  = x[ix.start + wpp[0]]
            y_  = iy.start + wpp[1]
            z_  = usable_data_orig[ wpp[0], wpp[1] ]

            X   = np.array( [ x_, y_, np.ones( len(x_) ) ] ).T
            mod = sm.OLS( z_, X )
            res = mod.fit()
            A, B, C = res.params
            ux  = x[ ix ]
            uy  = np.arange( iy.start, iy.stop)
            u_base  = compute_plane( A, B, C, ux, uy )
            if i == 0:
                usable_data -= u_base

        # TODO: replace LOW_PERCENTILE_BASE_FIN with a dynamically determined percentile
        kth = int( usable_data.size * LOW_PERCENTILE_BASE_FIN / 100 )
        zii     = np.argpartition( usable_data, kth, axis=None )
        zii_2d  = np.unravel_index( zii[0], usable_data.shape )
        kth_index = [ ix.start + zii_2d[0], iy.start + zii_2d[1] ]
        iff = kth_index[0]
        xff = x[iff]
        yff = kth_index[1]
        zff = data[ iff, yff ]
        # self.final_base_point = ( xff, yff, zff )
        C_  = zff - ( A * xff + B * yff )

        if debug:
            uv_elution  = ecurve_y[iy]
            e_curve     = ElutionCurve( uv_elution )
            debug_plot("Absorbance.solve_bottomplane_LPM debug", data, x, ix, iy, e_curve, smp_i, A, B, C_)

        self.params = ( A, B, C_ )

    def get_params(self):
        return self.params

class LambertBeerPlane:
    def __init__(self, data, x, ix, iy, ecurve_y, smp_i, istd_, debug=False):
        print('ix=', ix)
        print('iy=', iy)
        print('data.shape=', data.shape)
        print('data[ix, iy].shape=', data[ix,iy].shape)
        y   = np.arange( iy.start, iy.stop )
        uv_elution  = ecurve_y[iy]
        e_curve     = ElutionCurve( uv_elution )

        if False:
            from molass_legacy.Elution.CurveUtils import simple_plot
            print('iy.start, iy.stop=', iy.start, iy.stop)
            print('e_curve.peak_info=', e_curve.peak_info)
            print('e_curve.boundaries=', e_curve.boundaries)
            fig = plt.figure()
            ax = fig.gca()
            simple_plot(ax, e_curve, "LambertBeerPlane.__init__")
            plt.show()

        x_  = x[ix]
        M   = data[ix,iy]

        def obj_func( params ):
            A   = params['A']
            B   = params['B']
            C   = params['C']
            BP  = compute_plane( A, B, C, x_, y )
            M_  = M - BP

            P_list = []
            C_list = []
            start = 0
            for k, info in enumerate( e_curve.peak_info ):
                if k < len(e_curve.boundaries):
                    stop = e_curve.boundaries[k] - iy.start
                else:
                    stop = len(uv_elution) - iy.start

                # print([k], start, stop)
                if start >= stop:
                    # occured in 20170307
                    # is this ok?
                    continue

                peak_j  = int( info[1]+0.5 ) - iy.start
                P_list.append( M_[:,peak_j] )
                C_ = np.zeros(len(uv_elution))
                V_ = M_[istd_,start:stop]       # extract the corrected concentration
                C_[start:stop] = V_/np.max(V_)  # normalize to the max
                C_list.append(C_)

                start = stop

            P_ = np.array(P_list).T
            C_ = np.array(C_list)

            if False:
                from mpl_toolkits.mplot3d import Axes3D
                fig = plt.figure( figsize=( 10, 10 ) )
                ax = fig.add_subplot(111, projection='3d')
                ax.set_title("obj_func2 debug")
                for k, info in enumerate( e_curve.peak_info ):
                    peak_j  = int( info[1]+0.5 )
                    y_ = np.ones(len(x_))*peak_j
                    z_ = P_[:,k]
                    ax.plot(x_, y_, z_, color='green')
                    xs = np.ones(len(y))*x          # bug x
                    z_ = C_[k,:]*P_[istd_,k]
                    ax.plot(xs, y, z_)
                fig.tight_layout()
                plt.show()

            return ( np.dot(P_, C_) - M_ ).flatten()

        params = Parameters()
        params.add('A', value=0,   min=-1e-2,  max=+1e-2 )
        params.add('B', value=0,   min=-1e-2,  max=+1e-2 )
        params.add('C', value=0,   min=-1.0,   max=+1.0 )

        result  = minimize( obj_func, params, args=() )
        A   = result.params['A'].value
        B   = result.params['B'].value
        C   = result.params['C'].value

        if debug:
            debug_plot("Absorbance.solve_bottomplane_LB debug", data, x, ix, iy, e_curve, smp_i, A, B, C)

        self.params = (A, B, C)

    def get_params(self):
        return self.params



# coding: utf-8
"""

    MatrixFactorization.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""
import copy
import numpy as np
from scipy import optimize
from scipy.interpolate import UnivariateSpline
from mpl_toolkits.mplot3d import Axes3D
from molass_legacy.KekLib.SciPyCookbook import smooth
from ModeledData import ModeledData, simple_plot_3d
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data

MAX_ITERATION_FMIN_CG   = 10000

def C_smooth_penalty_impl(C):
    penalty = 0
    for i in range(C.shape[0]):
        y = C[i,:]
        j = np.arange(C.shape[1]-1)
        penalty += np.sum((y[j] - y[j+1])**2)
    return penalty

def deriv_C_smooth_penalty_impl(C):
    penalty_list = []
    for i in range(C.shape[0]):
        y = C[i,:]
        j = np.arange(C.shape[1]-1)
        penalty_list.append(np.hstack([2*(y[j]-y[j+1]), [0]]))
    return np.hstack(penalty_list)

class MatrixFactorization:
    def __init__(self, M, index, ecurve):
        self.M = M
        self.index = index
        self.ecurve = ecurve

        n_angs = M.shape[0]
        n_elns = M.shape[1]

        P_init = self.make_P_init()
        C_init = self.make_C_init()

        if True:
            self.debug_plot(P_init, C_init, M, "Initial State")

        rank = len(self.ecurve.peak_info) + 1

        self.M_ = get_denoised_data(M, rank=rank)
        self.P_init = P = P_init
        self.C_init = C = C_init

        for k in range(3):
            P_ = self.solve_P(P, C)
            C_ = self.solve_C(P_, C)
            P = P_
            C = C_

            if True:
                self.debug_plot(P, C, self.M_, "After Iteration %d" % k)

    def make_P_init(self):
        P_list = []
        for rec in self.ecurve.peak_info:
            k = rec[1]
            P_list.append(smooth(self.M[:,k]))
        return np.array(P_list).T

    def make_C_init(self):
        C_list = []
        start = 0
        ey = self.ecurve.sy
        for p, rec in enumerate(self.ecurve.peak_info):
            k = rec[1]
            if p < len(self.ecurve.boundaries):
                stop = self.ecurve.boundaries[p]
            else:
                stop = len(ey)
            y = ey[start:stop]
            y_ = np.hstack([ np.zeros(start), y/np.max(y), np.zeros(len(ey)-stop) ])
            C_list.append(y_)
            start = stop
        return np.array(C_list)

    def debug_plot(self, P, C, M, title):
        fig = plt.figure(figsize=(14,7))
        fig.suptitle(title)

        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122)

        simple_plot_3d(ax1, M, alpha=0.2)

        x = np.arange(M.shape[0])
        y = np.arange(M.shape[1])
        for p, rec in enumerate(self.ecurve.peak_info):
            k = rec[1]
            y_ = np.ones(len(x))*k
            z_ = P[:,p]
            ax1.plot(x, y_, z_, color='yellow')
            scale = np.max(z_)
            x_ = np.ones(len(y))*self.index
            z_ = C[p,:]*scale
            ax1.plot(x_, y, z_)
            ax2.plot(y, z_)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        plt.show()

    def solve_P(self, P, C):

        self.C = C
        P_ = self.M_ @ np.linalg.pinv(C)

        self.num_iterations = 0
        def num_iterations_count(x):
            self.num_iterations += 1

        ret = optimize.fmin_cg( self.compute_P_error, P_, fprime=self.compute_deriv_P_error,
                maxiter=MAX_ITERATION_FMIN_CG,
                full_output=True,
                callback=num_iterations_count )

        P_flat = ret[0]
        return P_flat.reshape( self.P_init.shape )

    def compute_P_error(self, P_flat):
        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C )

        # compute a squared Frobenius norms
        rec_error = np.linalg.norm( reconstructed - self.M_ )**2
        return rec_error

    def compute_deriv_P_error(self, P_flat):
        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C )

        # compute the derivative of the squared Frobenius norm
        # || P@C - M ||**2 ==> 2*( P@C - M )@C.T
        d_rec = 2 * np.dot( reconstructed - self.M_, self.C.T ).flatten()
        return d_rec

    def solve_C(self, P, C):

        self.P = P

        self.num_iterations = 0
        def num_iterations_count(x):
            self.num_iterations += 1

        ret = optimize.fmin_cg( self.compute_C_error, C, fprime=self.compute_deriv_C_error,
                maxiter=MAX_ITERATION_FMIN_CG,
                full_output=True,
                callback=num_iterations_count )

        C_flat = ret[0]
        return C_flat.reshape( self.C_init.shape )

    def compute_C_error(self, C_flat):
        C = C_flat.reshape( self.C_init.shape )
        reconstructed = np.dot( self.P, C )

        # compute a squared Frobenius norms
        recon_error = np.linalg.norm( reconstructed - self.M_ )**2
        smooth_penalty = C_smooth_penalty_impl(C)
        monot_penalty = self.compute_C_monot_penalty(C)

        return recon_error + smooth_penalty + monot_penalty

    def compute_deriv_C_error(self, C_flat):
        C = C_flat.reshape( self.C_init.shape )
        reconstructed = np.dot( self.P, C )

        # compute the derivative of the squared Frobenius norm
        # || P@C - M ||**2 ==> 2*( P@C - M )@C.T
        d_recon_error = 2 * np.dot( self.P.T, reconstructed - self.M_ ).flatten()
        d_smooth_penalty = deriv_C_smooth_penalty_impl(C)
        d_monot_penalty = self.compute_deriv_C_monot_penalty(C)

        return d_recon_error + d_monot_penalty

    def compute_C_monot_penalty(self, C):
        x = self.ecurve.x
        penalty = 0
        for i, rec in enumerate(self.ecurve.peak_info):
            top_j = rec[1]

            y = C[i,:]
            spline = UnivariateSpline(x, y, s=0, ext=3)
            dspline = spline.derivative(1)
            dy = dspline(x)
            dy[np.logical_and(dy > 0, x < top_j)] = 0
            dy[np.logical_and(dy < 0, x > top_j)] = 0
            penalty += np.sum(dy**2)
        return penalty

    def compute_deriv_C_monot_penalty(self, C):
        x = self.ecurve.x
        penalty = 0
        deriv_list =[]
        for i, rec in enumerate(self.ecurve.peak_info):
            top_j = rec[1]

            y = C[i,:]
            spline = UnivariateSpline(x, y, s=0, ext=3)
            dspline = spline.derivative(1)
            dy = dspline(x)
            dy[np.logical_and(dy > 0, x < top_j)] = 0
            dy[np.logical_and(dy < 0, x > top_j)] = 0
            deriv_list.append(2*dy)

        return np.hstack(deriv_list)



"""
    IFT/IftOptimizeImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.ModelUtils import compute_cy_list

def ift_optimize(M, E, qv, x, y, model, params_array, debug=True):
    cy_list = compute_cy_list(model, x, params_array)
    C = np.array(cy_list)
    # P = MC⁺
    
    Cinv = np.linalg.pinv(C)
    P = M @ Cinv
    


"""
    Simulative.DevelTest.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def devel_test_impl(editor):
    print("devel_test_impl")
    from importlib import reload
    import Batch.DataBridgeUtils
    reload(Batch.DataBridgeUtils)
    from molass_legacy.Batch.DataBridgeUtils import get_lrf_source_impl

    lrf_src = get_lrf_source_impl(editor)

    def moments_study():
        print("another_monopore_study")
        import Models.Stochastic.MomentsStudy
        reload(Models.Stochastic.MomentsStudy)
        from molass_legacy.Models.Stochastic.MomentsStudy import moments_study_impl
        moments_study_impl(lrf_src, debug=True)

    def show_num_plates_demo():
        import SecTheory.MartinSynge
        reload(SecTheory.MartinSynge)
        from SecTheory.MartinSynge import num_plates_demo
        num_plates_demo()

    def show_current_model_info():
        model = editor.get_current_model()
        modelname = editor.get_current_modelname()
        params_array = editor.get_current_params_array()
        # print("show_current_model_info", modelname)
        # print(params_array)
        return model, modelname, params_array

    def dispersive_pdf():
        import Models.Stochastic.DispersivePdfStudy
        reload(Models.Stochastic.DispersivePdfStudy)
        from molass_legacy.Models.Stochastic.DispersivePdfStudy import study_pdf
        study_pdf()

    def dispersive_time():
        import Models.Stochastic.DispersivePdf
        reload(Models.Stochastic.DispersivePdf)
        from molass_legacy.Models.Stochastic.DispersivePdf import timescale_proof
        timescale_proof()

    def dispersive_study(debug=False):
        import Models.Stochastic.DispersiveStudy
        reload(Models.Stochastic.DispersiveStudy)
        from molass_legacy.Models.Stochastic.DispersiveStudy import study
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(want_num_components=4, debug=False)
        info = show_current_model_info()
        study(lrf_src.xr_x, lrf_src.xr_y, lrf_src.baselines[1], lrf_src.model, lrf_src.xr_peaks[indeces,:], peak_rgs, qualities, props, curent_info=info, debug=debug)

    def test_stochastic_adapter(editor):
        print("test_stochastic_adapter")
        from importlib import reload
        import Selective.StochasticAdapter
        reload(Selective.StochasticAdapter)
        from Selective.StochasticAdapter import convert_to_stochastic_decomposition
        convert_to_stochastic_decomposition(editor, debug=True)

    extra_button_specs = [
            ("Guess Lnpore Params", lrf_src.guess_lnpore_params),
            ("Moments Study", moments_study),
            ("Moments Demo", lrf_src.moments_demo),
            ("Monopore Study", lrf_src.monopore_study),
            ("Num Plates Demo", show_num_plates_demo),
            ("Current Model Info", show_current_model_info),
            ("Dispersive PDF", dispersive_pdf),
            ("Dispersive TIme", dispersive_time),
            ("Dispersive Debug", lambda: dispersive_study(debug=True)),
            ("Dispersive Study", lambda: dispersive_study(debug=False)),
            ("Lnpore Study", lrf_src.lnpore_study),
            # ("Stochastic Adapter", lambda : test_stochastic_adapter(editor)),
            ] 

    lrf_src.draw(extra_button_specs=extra_button_specs)


"""
    Models.Stochastic.DispersiveLimits.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.Stochastic.ParamLimits import ParamsScaler, make_monopore_bounds

UNDEF_BOUND = np.nan, np.nan

class DispersiveParamsScaler(ParamsScaler):
    def __init__(self, egh_moments_list, num_scales, peak_rgs, unreliable_indeces,
                 poresize_bounds=None, fronting=False, N0_bound=None):
        self.num_scales = num_scales
        self.num_unreliables = len(unreliable_indeces)
        reliable_indeces = np.setdiff1d(np.arange(len(peak_rgs)), unreliable_indeces)
        if len(reliable_indeces) > 0:
            reliable_rgmax = np.max(peak_rgs[reliable_indeces])
        else:
            # as in 20161104/BL-6A/pH6
            reliable_rgmax = np.max(peak_rgs)
        if np.isnan(reliable_rgmax):
            # as in 20161104/BL-6A/pH6
            reliable_rgmax = 100

        eghM = egh_moments_list[0]
        t0_upper = eghM[0]

        temp_bounds = make_monopore_bounds(t0_upper, num_scales, self.num_unreliables, reliable_rgmax, 0)
        self.bounds = np.concatenate([temp_bounds[0:4], [UNDEF_BOUND], temp_bounds[4:]])
        self.bounds[0] =  np.array([200, 3000])     # N
        self.bounds[1] =  np.array([200, 2000])     # K = NT

        M = np.array(egh_moments_list)
        total_sigma = (M[-1][0] + np.sqrt(M[-1][1])) - ( M[0][0] - np.sqrt(M[0][1]) )
        print("total_sigma=", total_sigma)
        t0_lower = t0_upper - total_sigma
        if fronting:
            # as in 20230303/HasA
            t0_lower -= 2*total_sigma
            t0_upper -= 1*total_sigma

        self.bounds[2] = t0_lower, t0_upper
        if poresize_bounds is not None:
            self.bounds[3] = poresize_bounds
        tI_bounds = self.bounds[2] - total_sigma*np.array([2, 1])
        if tI_bounds[1] > 0:
            tI_bounds[1] = 0   # i.e., tI <= 0
        tI_bounds[0] = min(tI_bounds[0], tI_bounds[1] - 1)
        self.bounds[4] = tI_bounds

        print("self.bounds=", self.bounds )
        """
        bounds[0]   N
        bounds[1]   K = NT
        bounds[2]   x0+
        bounds[3]   poresize
        bounds[4]   tI
        bounds[5]   unreliable Rg1
        """
        self.set_scales()
        print("self.scales=", self.scales)


"""
    Elution.CurveUtils.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import logging
from bisect import bisect_right
import numpy as np
import matplotlib.patches as mpl_patches      # 'as patches' does not work properly
from itertools import combinations
import molass_legacy.KekLib.OurStatsModels as sm
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
from molass_legacy.KekLib.SciPyCookbook import smooth

PEAK_FIND_ALLOW     = 2.0       # > abs( 418.5 - 416.85 ) for 20180225

def is_included_in( x, x_list ):
    found = False
    for x_ in x_list:
        if abs( x_ - x ) < PEAK_FIND_ALLOW:
            found = True
            break
    return found

def get_around_slice( vector, val, num ):
    num_half    = num//2
    num_r       = num%2
    i = bisect_right( vector, val )
    return max( 0, i - num_half ), min( len(vector), i + num_half + num_r )

def get_xray_elution_vector( qvector, xray_array ):
    intensity_picking       = get_xray_picking()
    num_points_intensity    = get_setting( 'num_points_intensity' )
    slice_params = get_around_slice( qvector, intensity_picking, num_points_intensity )
    slice_ = slice( slice_params[0], slice_params[1] )
    # print( '===== slice_params=', slice_params, 'qvector[slice_]=', qvector[slice_] )
    return np.average( xray_array[:, slice_, 1], axis=1 ), slice_

def get_probably_corresponding_index( a_curve, x_curve ):
    a_size  = len(a_curve.x)
    x_size  = len(x_curve.x)

    a_peaks = np.array( [ info[1] for info in a_curve.peak_info ] )
    x_peaks = np.array( [ info[1] for info in x_curve.peak_info ] )

    num_a   = len(a_peaks)
    num_x   = len(x_peaks)
    index_a = list(range(num_a))
    index_x = list(range(num_x))

    if num_a == num_x:
        return [ index_a, index_x ], None

    if False:
        from CanvasDialog   import CanvasDialog

        def plot_func( fig ):
            ax1 = fig.add_subplot( 121 )
            ax2 = fig.add_subplot( 122 )

            ax1.plot( a_curve.y )
            ax2.plot( x_curve.y )

            def plot_peak( ax, curve ):
                for info in curve.peak_info:
                    peak = info[1]
                    x   = peak
                    y   = curve.spline( x )
                    ax.plot( x, y, 'o', color='red' )

            plot_peak( ax1, a_curve )
            plot_peak( ax2, x_curve )

            fig.tight_layout()
        dialog = CanvasDialog( "Debug: get_probably_corresponding_index", adjust_geometry=True )
        dialog.show( plot_func, figsize=(16, 8) )
        if not dialog.applied:
            from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
            debug_curve = ElutionCurve( x_curve.y, debug_plot=True )

    if num_a > num_x:
        reversed_ = False
        peaks1  = a_peaks
        peaks2  = x_peaks
        index1  = index_a
        index2  = index_x
        num1    = num_a
        num2    = num_x
        size1   = a_size
        size2   = x_size
    else:
        reversed_ = True
        peaks1  = x_peaks
        peaks2  = a_peaks
        index1  = index_x
        index2  = index_a
        num1    = num_x
        num2    = num_a
        size1   = x_size
        size2   = a_size

    def evaluate_mapping( index ):
        y   = [ 0, size1-1 ]
        x   = [ 0, size2-1 ]
        w   = [ 1, 1 ]
        for p1, p2 in zip( peaks1[index], peaks2 ):
            y.append( p1 )
            x.append( p2 )
            w.append( 10 )

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()
        return res

    # find minimum ssr result
    min_result  = None
    min_index   = None
    for c in combinations(index1, num2):
        index = list(c)
        res = evaluate_mapping( index )
        if min_result is None or res.ssr < min_result.ssr:
            min_result  = res
            min_index   = index

    intercept, slope = min_result.params
    if reversed_:
        return (index2, min_index), ( 1/slope, -intercept/slope )
    else:
        return (min_index, index2), ( slope, intercept )

SMOOTHING_WINDOW_LEN    = 10

def smoothed_curve_y(curve, y=None):
    if y is None:
        y = curve.y
    try:
        num_points = int(curve.get_primary_peak_num_points()/4) + 2
    except:
        # 20170426 why?
        logger = logging.getLogger(__name__)
        logger.warning("curve.get_primary_peak_num_points() failed.")
        num_points = SMOOTHING_WINDOW_LEN
    # num_points = 5 for 20190309_3
    # print('num_points=', num_points)
    window_len = min(SMOOTHING_WINDOW_LEN, num_points)
    sy = smooth(y, window_len=window_len)
    return sy

def make_smoothed_y(y, window_len=SMOOTHING_WINDOW_LEN):
    approximate_peak_num_points = len(y)/8
    num_points = int(approximate_peak_num_points/4) + 2
    # num_points = 5 for 20190309_3
    # print('num_points=', num_points)
    window_len = min(window_len, num_points)
    sy = smooth(y, window_len=window_len)
    return sy

MAJOR_PEAK_RATIO        = 0.03

def proof_plot( curve, parent, fig, in_folder=None, data_type='', fontsize=16, feature=False ):
    # TODO: refactoring to remove parent

    if in_folder is not None:
        elution = 'Elution' if data_type == '' else ' elution'
        fig.suptitle( data_type + elution + ' curve from folder ' + in_folder, fontsize=fontsize )

    ax1  = fig.add_subplot( 121 )
    ax2  = fig.add_subplot( 122 )

    ax1.set_title( "Recognized peaks", fontsize=fontsize )
    ax2.set_title( "Derivatives and roots for major peaks", fontsize=fontsize )

    x   = curve.x
    y   = curve.y
    spline  = curve.spline
    d1  = curve.d1
    d2  = curve.d2
    yˈ  = curve.d1y
    yˈˈ = d2(x)

    ptx = curve.peak_top_x
    ranges = curve.peak_info
    boundaries = curve.boundaries
    feet = np.array([[r[0], r[2]] for r in ranges]).flatten()

    ax1.plot( x, y, label='data' )
    ax1.plot( x, spline(x), label='spline' )
    ax1.plot( x, curve.feature_y, ':', label='feature_y' )
    # iknots = curve.knots[1:-1]
    # ax1.plot( iknots, spline(iknots), 'o', color='green', label='knots' )

    xmin, xmax = ax1.get_xlim()
    ax1.set_xlim(xmin, xmax)
    y_ = curve.min_y + curve.height * MAJOR_PEAK_RATIO
    ax1.plot( [xmin, xmax], [y_, y_], ':', color='red', label='peak recognition limit' )
    ax1.plot( ptx, spline(ptx), 'o', color='red', label='recongnized peak tops' )
    ax1.plot( feet, spline(feet), 'o', color='yellow', label='foot points' )

    if len(boundaries) > 0:
        ax1.plot( boundaries, spline(boundaries), 'o', color='green', label='recongnized peak boundaries' )

    if feature:
        ax1.plot( curve.feature.roots, spline(curve.feature.roots), 'o', color='yellow', label='inflections' )

    ax2.plot( x, yˈ, label='1st derivative' )
    ax2.plot( x, yˈˈ, label='2nd derivative' )
    ax2.plot( ptx, d1(ptx), 'o', color='red', label='recongnized peak tops' )
    if feature:
        ax2.plot( curve.feature.roots, d1(curve.feature.roots), 'o', color='yellow', label='inflections' )

    ymin, ymax = ax1.get_ylim()
    for range_ in ranges:
        f = range_[0]
        t = range_[2]
        p = mpl_patches.Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.1,
            )
        ax1.add_patch( p )

    ax1.legend(fontsize=fontsize)
    ax2.legend(fontsize=fontsize)
    fig.tight_layout()
    if in_folder is not None:
        fig.subplots_adjust( top=0.88 )

def simple_plot(ax, curve, title=None, legend=True, color=None, major_peaks=False, boundaries=None, spline=False ):
    x = curve.x
    y = curve.y
    if major_peaks:
        top = [ info[1] for info in curve.get_major_peak_info() ]
        bnd = curve.get_major_valley_bottoms(top)
    else:
        top = curve.peak_top_x
        bnd = curve.boundaries if boundaries is None else boundaries
    if title is not None:
        ax.set_title(title, fontsize=16)
    ax.plot(x, y, color=color, label='data')
    ax.plot(top, curve.spline(top), 'o', color='red', label='peak tops')
    if len(bnd) > 0:
        ax.plot(bnd, curve.spline(bnd), 'o', color='green', label='valley bottoms')

    if spline:
        ax.plot(x, curve.sy, ':', color='green', label='spline')

    if legend:
        ax.legend()

# TODO: remove duplication of this code
def rotate( th, wx, wy ):
    k = -1 if th > 0 else 0
    cx  = wx[k]
    cy  = wy[k]
    wx_ = wx - cx
    wy_ = wy - cy
    c   = np.cos( th )
    s   = np.sin( th )
    return cx + ( wx_*c - wy_* s ), cy + ( wx_*s + wy_* c )

MIN_NUM_POINTS  = 5

def find_rotated_extreme_arg(curve, start, stop, extremun_flag):
    slice_ = slice(start, stop)

    x = curve.x[slice_]
    y = curve.sy[slice_]    # use the smoothed

    assert len(x) >= MIN_NUM_POINTS

    x1, x2 = x[[0, -1]]
    y1, y2 = y[[0, -1]]

    xscale = 1/abs(x2 - x1)
    yscale = 1/abs(y2 - y1)

    x_ = x * xscale
    y_ = y * yscale

    y_end_ratio = (y_[-1] - y_[0])/curve.height     # better be nearest peak height

    if abs(y_end_ratio) < 0.5:
        rx, ry = x_, y_
    else:
        rotate_sign = -1 if y_end_ratio > 0 else 1
        angle = rotate_sign* np.pi/4
        rx, ry = rotate(angle, x_, y_)

    if extremun_flag < 0:
        pnn = np.argpartition(ry, MIN_NUM_POINTS)
        n = int(np.average(pnn[0:5]) + 0.5)
    else:
        k = len(ry) - MIN_NUM_POINTS
        pnn = np.argpartition(ry, k)
        n = int(np.average(pnn[k:]) + 0.5)

    ret_n = start + n

    if False:
        import molass_legacy.KekLib.DebugPlot as plt

        plt.push()
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
        ax1, ax2 = axes

        fig.suptitle("find_rotated_extreme_arg")
        simple_plot(ax1, curve)
        ax1.plot(curve.x[ret_n], curve.y[ret_n], 'o', color='yellow')

        x_ = curve.x[start:stop]
        y_ = curve.y[start:stop]
        ax2.plot(rx, ry)
        ax2.plot(rx[n], ry[n], 'o', color='yellow')
        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        plt.show()
        plt.pop()

    return ret_n



"""
PythonでGmailを送信する
https://qiita.com/eito_2/items/ef77e44955e43f31ba78

Gmail（Google）アカウント – アプリパスワードの生成（作成）
https://pc-karuma.net/google-account-generate-app-password/
"""
import os
import smtplib
from email.mime.text import MIMEText
from email.utils import formatdate

def send_notify_gmail(subject, body):
    sendAddress = os.environ.get("MY_GMAIL_ADDRESS")
    if sendAddress is None:
        return

    password = os.environ.get("MY_GMAIL_PASSWORD")
    fromAddress = sendAddress
    toAddress = sendAddress

    # SMTPサーバに接続
    smtpobj = smtplib.SMTP('smtp.gmail.com', 587)
    smtpobj.starttls()
    smtpobj.login(sendAddress, password)

    # メール作成
    msg = MIMEText(body)
    msg['Subject'] = subject
    msg['From'] = fromAddress
    msg['To'] = toAddress
    msg['Date'] = formatdate()

    # 作成したメールを送信
    smtpobj.send_message(msg)
    smtpobj.close()

if __name__ == '__main__':
    subject = "Test"
    body = "test body"
    send_notify_gmail(subject, body)



# coding: utf-8
"""
    ExtrapolationProof.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import re
import copy
import numpy as np
from OurStatsModels import WLS, add_constant
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from ScrolledFrame import ScrolledFrame
import molass_legacy.KekLib.DebugPlot as plt
from SerialTestUtils import get_tttt_data_objects, debug_plot_runner
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from SvdDenoise import get_denoised_data
from SmoothFactorizer import SmoothFactorizer
from SerialTestUtils import get_tk_root
from DataUtils import get_pytools_folder, serial_folder_walk, cut_upper_folders
from SerialDataUtils import get_mtd_filename
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
from SerialAtsasTools import AlmergeExecutor

class ZxSolver:
    def plot_func(self):
        q   = self.q
        P   = self.P
        AB  = self.AB
        ABe = self.ABe

        fig = plt.figure()
        ax2 = fig.gca()
        ax2t = ax2.twinx()

        ax2.plot(q, np.log10(P[:,0]), label='A-matrix')
        ax2t.plot(q, P[:,1], label='B-matrix')
        # ax2.plot(q, np.log10(e[:,0]), label='A-matrix error')
        ax2.plot(q, np.log10(AB[:,0]), label='A-iter')
        # ax2.plot(q, ABe[:,0]), label='A-iter error')
        ax2t.plot(q, AB[:,1], label='B-iter')
        fig.tight_layout()
        plt.show()

    def plot(self):
        debug_plot_runner(self.plot_func)

    def plot_A(self, ax, label):
        q   = self.q
        P   = self.P
        ax.plot(q, np.log10(P[:,0]), label=label)

    def plot_B(self, ax, label):
        q   = self.q
        P   = self.P
        ax.plot(q, P[:,1], label=label)

class MpInverse(ZxSolver):
    def __init__(self, q, M, E, c):
        self.q  = q
        self.M  = M
        self.E  = E
        self.c  = c

    def solve(self, with_const=False):
        c_ = self.c
        c_list = [c_, c_**2]
        if with_const:
            c_list.append( np.ones(len(c_)) )

        self.C = np.array( c_list )
        self.Cpinv = np.linalg.pinv(self.C)
        self.P = np.dot(self.M, self.Cpinv)
        self.e = np.sqrt( np.dot(self.E**2, self.Cpinv**2) )
        print( self.P.shape, self.e.shape )

class IteratedWLS(ZxSolver):
    def __init__(self, q, M, E, c):
        self.q  = q
        self.M  = M
        self.E  = E
        self.c  = c
        c_ = c
        self.C = np.array( [c_, c_**2] )

    def solve(self, weighted=True, with_const=False):
        AB_list = []
        ab_error_list = []
        for i in range(self.M.shape[0]):
        # for i in [10]:
            Mi = np.array( [ self.M[i,:] ] )
            # print(M.shape, C.shape)
            if True:
                e_ = np.array( [ self.E[i,:] ] ).flatten()
                e2 = e_**2
                if weighted:
                    w = 1/e2
                else:
                    w = None
                x = Mi.flatten()
                if with_const:
                    X = add_constant(x)
                else:
                    X = x
                model = WLS(x, self.C.T, weights=w)
                result  = model.fit()
                Pi = result.params
                ab_error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
                ab_error_list.append( ab_error )
            else:
                Pi = np.dot(Mi, self.Cpinv)
            # print( Pi )
            AB_list.append( Pi.flatten() )

        self.P = np.array(AB_list)
        self.e = np.array(ab_error_list)
        print( self.P.shape, self.e.shape )

class NormMinizerSmoothed(ZxSolver):
    def __init__(self, q, M, E, c):
        self.q  = q
        self.M  = M
        self.E  = E
        self.c  = c

    def solve(self, apositive=False, smoothed=True):
        factorizer = SmoothFactorizer(self.M, self.E, self.c)
        factorizer.solve(apositive=apositive, smoothed=smoothed)
        self.P = factorizer.P

class AlMergeSolver:
    def __init__(self, q, M, E, c):
        self.q  = q
        self.M  = M
        self.E  = E
        self.c  = c
        self.temp_folder = get_setting('temp_folder') + '/almerge'
        clear_dirs_with_retry( [self.temp_folder] )

    def solve(self):
        # write files into temp_folder
        files = []
        for j in range(self.M.shape[1]):
            path = self.temp_folder + "/ALMERGE_IN_%03d.dat" % j
            files.append(path)
            data = np.array([self.q, self.M[:,j], self.E[:,j]]).T
            np.savetxt(path, data)

        executor = AlmergeExecutor()
        indeces = np.arange(len(files))
        out_file =  self.temp_folder + '/almerge_out.dat'
        result = executor.execute(self.c, files, indeces, out_file)
        self.A = result.exz_array[:,1]
        self.P = self.A.reshape( (self.A.shape[0], 1) )

    def scale(self, other95):
        this95 = np.percentile(self.A, 95)
        self.A = self.A * (other95/this95)
        self.P = self.A.reshape( (self.A.shape[0], 1) )

    def copy(self):
        clone = AlMergeSolver(self.q, self.M, self.E, self.c)
        clone.A = copy.deepcopy(self.A)
        clone.P = copy.deepcopy(self.P)
        return clone

    def plot_A(self, ax, label):
        q   = self.q
        ax.plot(q, np.log10(self.A), label=label)

    def plot_B(self, ax, label):
        pass

def create_plot_specs(in_folder, uv_folder, range_callback, range_=None):

    _, sd, mapper = get_tttt_data_objects(in_folder, uv_folder, analysis_copy=True, adjust=True )
    c_all = mapper.get_conc_vector( 5 )
    sd.apply_baseline_correction( mapper.get_mapped_info() )

    spec_list = []

    if range_ is None:
        for pno, range_ in enumerate(mapper.get_int_ranges()):
            print(range_)
            lower, middle, upper = range_
            for ad, ft in enumerate([ [lower, middle], [middle, upper] ]):
                f, t = ft
                spec = range_callback(in_folder, sd, pno, ad, (f, t), c_all)
                spec_list.append(spec)
    else:
        pno, ad, range__ = range_
        spec = do_a_range(in_folder, sd, pno, ad, range__, c_all)
        spec_list.append(spec)

    return spec_list

def do_a_range(in_folder, sd, pno, ad, range_, c_all):
    eslice = slice(*range_)

    q   = sd.intensity_array[0,:,0]
    D   = sd.intensity_array[eslice,:,1].T
    E   = sd.intensity_array[eslice,:,2].T
    c = c_all[eslice]

    solver0 = MpInverse(q, D, E, c)
    solver0.solve()
    solver1 = MpInverse(q, D, E, c)
    solver1.solve(with_const=True)
    # solver2 = IteratedWLS(q, D, E, c)
    # solver2.solve(weighted=False)
    solver3 = IteratedWLS(q, D, E, c)
    solver3.solve(weighted=True)
    solver4 = NormMinizerSmoothed(q, D, E, c)
    solver4.solve(smoothed=False)
    solver5 = NormMinizerSmoothed(q, D, E, c)
    solver5.solve(smoothed=False)
    solver6 = NormMinizerSmoothed(q, D, E, c)
    solver6.solve(smoothed=False, apositive=True)

    data_name = in_folder + "/pk%d_%s_*.dat" % (pno+1, 'asc' if ad == 0 else 'dsc')
    return data_name, [
            # [ [ "MP-inverse",  solver0], [ "iterated-OLS", solver2 ] ],
            [ [ "MP-inverse",  solver0], [ "iterated-WLS", solver3 ] ],
            [ [ "MP-inverse",  solver0], [ "MP-inverse-with-1",  solver1] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min", solver4 ] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min-smooth", solver5] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min-A_positive", solver6] ],
            ]

def do_a_range_denoise(in_folder, sd, pno, ad, range_, c_all):
    eslice = slice(*range_)

    q   = sd.intensity_array[0,:,0]
    D   = sd.intensity_array[eslice,:,1].T
    D2_  = get_denoised_data(D, rank=2)
    D3_  = get_denoised_data(D, rank=3)
    D4_  = get_denoised_data(D, rank=4)
    D5_  = get_denoised_data(D, rank=5)

    E   = sd.intensity_array[eslice,:,2].T
    c = c_all[eslice]

    solver0 = MpInverse(q, D, E, c)
    solver0.solve()
    solver1 = IteratedWLS(q, D, E, c)
    solver1.solve(weighted=False)
    solver5 = MpInverse(q, D5_, E, c)
    solver5.solve()
    solver4 = MpInverse(q, D4_, E, c)
    solver4.solve()
    solver3 = MpInverse(q, D3_, E, c)
    solver3.solve()
    solver2 = MpInverse(q, D2_, E, c)
    solver2.solve()

    data_name = in_folder + "/pk%d_%s_*.dat" % (pno+1, 'asc' if ad == 0 else 'dsc')
    return data_name, [
            [ [ "MP-inverse",  solver0], [ "iterated-WLS",  solver1] ],
            [ [ "MP-inverse",  solver0], [ "MP-inverse (rank 5 denoised)",  solver5, "rank=5"] ],
            [ [ "MP-inverse",  solver0], [ "MP-inverse (rank 4 denoised)",  solver4, "rank=4"] ],
            [ [ "MP-inverse",  solver0], [ "MP-inverse (rank 3 denoised)",  solver3, "rank=3"] ],
            [ [ "MP-inverse",  solver0], [ "MP-inverse (rank 2 denoised)",  solver2, "rank=2"] ],
            ]

def do_a_range_definitive(in_folder, sd, pno, ad, range_, c_all):
    eslice = slice(*range_)

    q   = sd.intensity_array[0,:,0]
    D   = sd.intensity_array[eslice,:,1].T
    D3_  = get_denoised_data(D, rank=3)
    E   = sd.intensity_array[eslice,:,2].T
    c = c_all[eslice]

    solver0 = MpInverse(q, D3_, E, c)
    solver0.solve()
    solver1 = IteratedWLS(q, D3_, E, c)
    solver1.solve(weighted=True)
    solver2 = NormMinizerSmoothed(q, D3_, E, c)
    solver2.solve(smoothed=False, apositive=False)
    solver3 = NormMinizerSmoothed(q, D3_, E, c)
    solver3.solve(smoothed=True, apositive=False)
    solver4 = NormMinizerSmoothed(q, D3_, E, c)
    solver4.solve(smoothed=False, apositive=True)
    solver5 = NormMinizerSmoothed(q, D3_, E, c)
    solver5.solve(smoothed=True, apositive=True)

    data_name = in_folder + "/pk%d_%s_*.dat" % (pno+1, 'asc' if ad == 0 else 'dsc')
    return data_name, [
            [ [ "MP-inverse",  solver0], [ "iterated-WLS",  solver1] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min",  solver2] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min-smooth)",  solver3] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min-A_positive)",  solver4] ],
            [ [ "MP-inverse",  solver0], [ "Norm-min-smooth-A_positive)",  solver5] ],
            ]

def do_a_range_almerge(in_folder, sd, pno, ad, range_, c_all):
    eslice = slice(*range_)

    q   = sd.intensity_array[0,:,0]
    D   = sd.intensity_array[eslice,:,1].T
    D3_  = get_denoised_data(D, rank=3)
    E   = sd.intensity_array[eslice,:,2].T
    c = c_all[eslice]

    solver0 = MpInverse(q, D3_, E, c)
    solver0.solve()
    solver1 = AlMergeSolver(q, D3_, E, c)
    solver1.solve()
    solver3 = solver1.copy()
    solver5 = solver1.copy()
    solver1.scale(np.percentile(solver0.P[:,0], 95))
    solver2 = NormMinizerSmoothed(q, D3_, E, c)
    solver2.solve(smoothed=True, apositive=False)
    solver3.scale(np.percentile(solver2.P[:,0], 95))
    solver4 = NormMinizerSmoothed(q, D3_, E, c)
    solver4.solve(smoothed=True, apositive=True)
    solver5.scale(np.percentile(solver4.P[:,0], 95))

    data_name = in_folder + "/pk%d_%s_*.dat" % (pno+1, 'asc' if ad == 0 else 'dsc')
    return data_name, [
            [ [ "MP-inverse",  solver0], [ "ALMERGE",  solver1] ],
            [ [ "Norm-min-smooth",  solver2], [ "ALMERGE",  solver3] ],
            [ [ "Norm-min-smooth-A_positive",  solver4], [ "ALMERGE",  solver5] ],
            ]

def compute_nrmsd(max_y, y1, y2):
    return np.sqrt( np.average( (y1 - y2)**2 ) )/max_y*100

class ProofViewer(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.applied = False

    def show(self, data_name, plot_specs, save_images=True, auto=True, show_bq=True):
        self.data_name = cut_upper_folders(data_name)
        self.plot_specs = plot_specs
        self.image_folder = get_setting('temp_folder') + '/images'
        self.save_images = save_images
        self.auto = auto
        self.show_bq = show_bq
        title   = "ProofViewer"
        Dialog.__init__( self, self.parent, title )

    def body( self, body_frame ):

        sframe = ScrolledFrame(body_frame )
        sframe.pack()
        toolbar_frame = Tk.Frame(body_frame)
        toolbar_frame.pack()

        cframe = sframe.interior
        nrows = max(2, len(self.plot_specs))
        ncols = 2 if self.show_bq else 1

        height = 2 * nrows if self.show_bq else 3 * nrows
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(16, height))
        fig.tight_layout()

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.toolbar = NavigationToolbar( self.mpl_canvas, toolbar_frame )
        self.toolbar.update()

        self.fig = fig
        self.axes = axes

        fig.suptitle("Comparison of Extrapolation Methods using " + self.data_name, fontsize=20)

        t_ymin1 = []
        t_ymax1 = []
        t_ymin2 = []
        t_ymax2 = []
        for i, spec_row in enumerate(self.plot_specs):
            if self.show_bq:
                ax1 = axes[i,0]
                ax2 = axes[i,1]
            else:
                ax1 = axes[i]
            ax1.set_title("A(q)")
            ax1.set_ylabel('$Log_{10}$( Intensity )')
            if self.show_bq:
                ax2.set_title("B(q)")
                ax2.set_ylabel('$Linear$( Intensity )')

            for spec in spec_row:
                label, solver = spec[0:2]
                solver.plot_A(ax1, label)
                if self.show_bq:
                    solver.plot_B(ax2, label)

                ymin, ymax = ax1.get_ylim()
                t_ymin1.append(ymin)
                t_ymax1.append(ymax)
                if self.show_bq:
                    ymin, ymax = ax2.get_ylim()
                    t_ymin2.append(ymin)
                    t_ymax2.append(ymax)

            ax1.legend()
            if self.show_bq:
                ax2.legend()

        f_ymin1 = np.percentile( t_ymin1, 5 )
        f_ymax1 = np.percentile( t_ymax1, 100 )
        if self.show_bq:
            f_ymin2 = np.percentile( t_ymin2, 10 )
            f_ymax2 = np.percentile( t_ymax2, 90 )
        for i, spec_row in enumerate(self.plot_specs):
            if self.show_bq:
                ax1 = axes[i,0]
                ax2 = axes[i,1]
            else:
                ax1 = axes[i]
            ax1.set_ylim(f_ymin1, f_ymax1)
            if self.show_bq:
                ax2.set_ylim(f_ymin2, f_ymax2)

            if self.show_bq:
                def get_txext_xy_ax2():
                    xmin, xmax = ax2.get_xlim()
                    ymin, ymax = ax2.get_ylim()
                    tx = xmin*0.9 + xmax*0.1
                    if (ymin + ymax)/2 < 0:
                        ty = ymin*0.8 + ymax*0.2
                    else:
                        ty = ymin*0.4 + ymax*0.6
                    return tx, ty

            P_list = []
            text_drawn = False
            for spec in spec_row:
                solver = spec[1]
                P_list.append(solver.P)
                if len(spec) > 2:
                    text = spec[2]
                else:
                    text = None
                if text is not None:
                    text_drawn = True
                    if self.show_bq:
                        tx, ty = get_txext_xy_ax2()
                        ax2.text(tx, ty, text, ha='left', fontsize=30, alpha=0.3)

            max_y = np.percentile(P_list[0][:,0], 95)
            if not text_drawn and self.show_bq:
                if P_list[1].shape[1] > 1:
                    b_nrsmd = compute_nrmsd(max_y, P_list[0][:,1], P_list[1][:,1] )
                    text = "nRMSD=%5.2f%%" % (b_nrsmd)
                    tx, ty = get_txext_xy_ax2()
                    ax2.text(tx, ty, text, ha='left', fontsize=30, alpha=0.3)

            a_nrmsd = compute_nrmsd(max_y, P_list[0][:,0], P_list[1][:,0] )
            xmin, xmax = ax1.get_xlim()
            ymin, ymax = ax1.get_ylim()
            tx = xmin*0.8 + xmax*0.2
            ty = ymin*0.3 + ymax*0.7
            text = "nRMSD=%5.2f%%" % (a_nrmsd)
            ax1.text(tx, ty, text, ha='left', fontsize=30, alpha=0.3)
        if self.show_bq:
            top = 0.75 + (nrows-2)*0.05     # i.e., 0.75 if nrows == 2, 0.9 if nrows == 5
        else:
            top = 0.85 + (nrows-2)*0.05     # i.e., 0.85 if nrows == 2, 0.9 if nrows == 3
        fig.subplots_adjust( top=top, left=0.05, wspace=0.18 )

        if self.save_images:
            self.save_image()

    def apply(self):
        self.applied = True

    def save_image(self):
        name_ = self.data_name.replace('_*.dat', '')
        filename = name_.replace('/', '-') + '.png'
        path = os.path.join( self.image_folder, filename ).replace('/', '\\')
        print('save_image: path=', path)
        self.fig.savefig( path )
        if self.auto:
            self.after(1000, self.ok)

def survey_for_all_data(restart=None, auto=True, callback=do_a_range, show_bq=True):
    root = get_tk_root()
    started = False

    def do_a_folder(in_folder, uv_folder, plot):
        nonlocal started
        mtd_file = get_mtd_filename(uv_folder)
        if mtd_file is not None:
            return True, None

        if restart is not None:
            if not started:
                if in_folder.find(restart) < 0:
                    return True, None
                else:
                    started = True

        print(in_folder)
        try:
            spec_list = create_plot_specs(in_folder, uv_folder, callback)
            for spec in spec_list:
                data_name, plot_specs = spec
                viewer = ProofViewer(root)
                viewer.show(data_name, plot_specs, auto=auto, show_bq=show_bq)
                if not viewer.applied:
                    break

            return viewer.applied, None
        except:
            print('create_plot_specs failed for', in_folder )
            return True, None

    pytools = get_pytools_folder()
    data_folder = pytools + '/Data'
    print( data_folder )
    serial_folder_walk( data_folder, do_a_folder )
    


"""
    SecTheory.Sepsey2014b.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import basinhopping
import molass_legacy.KekLib.DebugPlot as plt

def ng(rg, rp, nperm, me):
    rho = rg/rp
    return nperm*(1 - rho)**me

def tg(rg, rp, tperm, mp):
    rho = rg/rp
    return tperm*(1 - rho)**mp

SQRT_2PI = np.sqrt(2*np.i)

def sepsey_phi_19(w, rp):
    pass

def demo():
    from .SecCF import simple_phi, moving_zone_phi
    from .SecPDF import c, FftInvPdf

    t = np.arange(300)

    npi = 10
    tpi = 10
    simple_pdf = FftInvPdf(simple_phi)

    moving_pdf = FftInvPdf(moving_zone_phi)
    N0 = 6
    t0 = 6

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Proof of Numerical Inversion", fontsize=20)
        ax1.set_title("Simple", fontsize=16)
        ax1.plot(t, c(t, npi, tpi), label="formula")
        ax1.plot(t, simple_pdf(t, npi, tpi), label="numerical")
        ax1.legend()
        ax2.set_title("Moving Zone considered", fontsize=16)
        ax2.plot(t, moving_pdf(t, npi, tpi, N0, t0), label="numerical")
        ax2.legend()
        fig.tight_layout()
        plt.show()



# coding: utf-8
"""
    CurveDecomposer.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from molass_legacy.Models.ElutionCurveModels import EGH
from molass_legacy.ElutionDecomposer import ElutionDecomposer

def decompose(ecurve, model=EGH()):
    x = ecurve.x
    y = ecurve.y
    decomposer = ElutionDecomposer(ecurve, x, y, model=model, retry_valley=True, deeply=True)
    return decomposer.fit_recs



"""

    SerialAnalyzer.Tester.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""

import os
import re
import queue
import numpy        as np
from time           import sleep
from datetime import datetime
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.SerialAnalyzer.DataUtils import serial_folder_walk, mct_folder_walk
from molass_legacy.KekLib.MessageBoxUtils import reply_messagebox, window_exists
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry, Struct
from molass_legacy._MOLASS.SerialSettings import clear_settings, get_setting, set_setting, DF_PATH_LENGTH, DF_PATH_LENGTH_OLD
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy.Test.TesterLogger import *

success_rg_re   = re.compile( r'Rg from (\S+) is estimated to be (\S+) with quality (\S+)\.' )
failure_rg_re   = re.compile( r'Rg estimation failed in \S+\.' )
pytools_ver_rg  = re.compile( r'pytools-\w+-develop' )
final_std_diff_str  = 'final_std_diff='

INVESTIGATE_ABSORBANCE_ANOMALY  = False
MINIMUM_DATA_NAMES = [
    'pH6',          # peak consistency when noisy (one peak)
    'pH7',          # peak consistency when noisy (two peaks)
    '20190315_3',   # No decomposition
    '20191006_proteins5',   # many peaks, 10-sigma rule covering
    '20191031',     # use_xray_conc
    'sample_data',  # standard
    '20190522',     # MCT-SAXS
    ]

PATH_LENGTH_DICT = {
    'pH6':DF_PATH_LENGTH_OLD,
    'pH7':DF_PATH_LENGTH_OLD,
    '20190315_3':DF_PATH_LENGTH_OLD,
    '20191006_proteins5':DF_PATH_LENGTH,
    '20191031':DF_PATH_LENGTH,
    'sample_data':DF_PATH_LENGTH_OLD,
    '20190522':DF_PATH_LENGTH_OLD,
    }

fromlog_rg_list = None

def compare_tester_rg( i, rg, quality_boundary ):
    if fromlog_rg_list is None:
        return

    if i >= len( fromlog_rg_list ):
        put_to_tester_log_queue( 'No %d-th result to compare\n' % (i) )
        return

    from_rec = fromlog_rg_list[i]
    if from_rec is None:
        from_rg         = None
        from_quality    = None
    else:
        from_rg         = from_rec[0]
        from_quality    = from_rec[1]
    if from_rg is None:
        if rg is None:
            pass
        else:
            put_to_tester_log_queue( 'got new Rg(%.3g) from %d-th file\n' % (rg, i) )
    else:
        if rg is None:
            put_to_tester_log_queue( 'got None Rg from %d-th file\n' % i )
        else:
            if from_rg > 0:
                diff_ratio = abs( rg - from_rg ) / from_rg
                if from_quality > quality_boundary:
                    allow_diff  = 0.01
                else:
                    allow_diff  = 0.1
                if diff_ratio > allow_diff:
                    put_to_tester_log_queue( 'got different Rg(%.3g != %.3g) from %d-th file\n' % (rg, from_rg, i) )
            else:
                pass

class Tester:
    def __init__( self, app ):
        self.app   = app

    def test_all( self, test_folders, result_folder=None,
                    report_folder=None, report_subfolder=None, filter_str=None, start_str=None,
                    test_patterns=[0,1,2],
                    minimum_only=False,
                    ds_filter=None,
                    mapping_only=False,
                    debug_trial=False,
                    save_outline_figures=False,
                    save_3d_figures=False,
                    save_baseline_figures=False,
                    save_mapping_figures=False,
                    save_cdi_figures=False,
                    save_decomp_figures=False,
                    save_preview_figures=False,
                    save_preview_results=False,
                    save_preview_cdi_figures=False,
                    save_peakeditor_figures=False,
                    shutdown_machine=False,
                    to_compare_logfolder=None,
                    v2_analysis_folders=None,
                    ):
        global log_fh, tester_log_queue
        tester_log_queue = queue.Queue()
        self.test_folders   = test_folders
        self.filter_str     = filter_str
        if start_str is not None and len( start_str.replace( ' ', '' ) ) == 0:
            start_str = None
        self.start_str      = start_str
        # self.start_str      = '20161217'
        self.minimum_only   = minimum_only
        self.ds_filter      = ds_filter
        self.mapping_only   = mapping_only
        self.allow_mct_folder = 0
        self.debug_trial    = debug_trial
        self.save_outline_figures   = save_outline_figures
        self.save_3d_figures        = save_3d_figures
        self.save_baseline_figures  = save_baseline_figures
        self.save_mapping_figures   = save_mapping_figures
        self.save_cdi_figures       = save_cdi_figures
        self.save_decomp_figures    = save_decomp_figures
        self.save_preview_figures   = save_preview_figures
        self.save_preview_results   = save_preview_results
        self.save_preview_cdi_figures = save_preview_cdi_figures
        self.save_peakeditor_figures = save_peakeditor_figures
        self.shutdown_machine = shutdown_machine
        self.to_compare_logfolder = to_compare_logfolder
        self.v2_analysis_folders = v2_analysis_folders
        self.diff_count     = 0
        self.error_count    = 0
        self.mdiff_count    = 0
        self.test_patterns  = test_patterns
        if result_folder is None:
            this_dir = os.path.dirname( os.path.abspath( __file__ ) )
            home_dir = this_dir + '/../..'
        else:
            home_dir = result_folder
        self.log    = open_tester_log( home_dir )

        if report_folder is None:
            report_folder   = os.path.abspath( home_dir + '/reports' ).replace( '\\', '/' )

        self.outline_image_folder = None
        self.threed_image_folder = None
        self.baseline_image_folder = None
        self.mapping_image_folder = None
        self.decomp_image_folder_xray = None
        self.decomp_image_folder_uv = None
        self.preview_image_folder = None
        self.qmm_image_folder = None
        self.fulloptinit_image_folder = None
        self.result_folder = result_folder

        dirs_to_clear = [ report_folder ]

        if self.save_outline_figures:
            self.outline_image_folder   = os.path.abspath( home_dir + '/images-outline' ).replace( '\\', '/' )
            dirs_to_clear.append( self.outline_image_folder )
        if self.save_3d_figures:
            self.threed_image_folder   = os.path.abspath( home_dir + '/images-3d' ).replace( '\\', '/' )
            dirs_to_clear.append( self.threed_image_folder )
        if self.save_baseline_figures:
            self.baseline_image_folder   = os.path.abspath( home_dir + '/images-baseline' ).replace( '\\', '/' )
            dirs_to_clear.append( self.baseline_image_folder )
        if save_mapping_figures:
            self.mapping_image_folder   = os.path.abspath( home_dir + '/images-mapping' ).replace( '\\', '/' )
            dirs_to_clear.append( self.mapping_image_folder )
        if self.save_cdi_figures:
            self.cdi_image_folder   = os.path.abspath( home_dir + '/images-cdi' ).replace( '\\', '/' )
            dirs_to_clear.append( self.cdi_image_folder )
        if save_decomp_figures:
            self.decomp_image_folder_xray   = os.path.abspath( home_dir + '/images-decomp-xray' ).replace( '\\', '/' )
            dirs_to_clear.append( self.decomp_image_folder_xray )
            self.decomp_image_folder_uv = os.path.abspath( home_dir + '/images-decomp-uv' ).replace( '\\', '/' )
            dirs_to_clear.append( self.decomp_image_folder_uv )
        if save_preview_figures:
            self.preview_image_folder   = os.path.abspath( home_dir + '/images-preview' ).replace( '\\', '/' )
            dirs_to_clear.append( self.preview_image_folder )
        if self.save_preview_cdi_figures:
            self.preview_cdi_image_folder = os.path.abspath( home_dir + '/images-preview-cdi' ).replace( '\\', '/' )
            dirs_to_clear.append( self.preview_cdi_image_folder )
        if self.save_peakeditor_figures:
            self.peakeditor_image_folder = os.path.abspath( home_dir + '/images-peakeditor' ).replace( '\\', '/' )
            dirs_to_clear.append( self.peakeditor_image_folder )

        if self.test_patterns[0] in [33333]:
            self.srr_folder = os.path.abspath( home_dir + '/srr' ).replace( '\\', '/' )
            dirs_to_clear.append( self.srr_folder )

        if self.test_patterns[0] in [4]:
            print("home_dir=", home_dir)
            survey_file = os.path.abspath( home_dir + '/mapping_params.csv' )
            self.survey_fh = open(survey_file, "w")

        if self.test_patterns[0] in [9, 10]:
            self.fulloptinit_image_folder = os.path.abspath( home_dir + '/images-fullopt_init' ).replace( '\\', '/' )
            dirs_to_clear.append( self.fulloptinit_image_folder )

        if self.test_patterns[0] >= 20 and self.test_patterns[0] < 30:
            self.make_v2_results_dict()

        while True:
            clear_dirs_with_retry( dirs_to_clear )
            sleep( 1 )
            if os.path.exists( report_folder ):
                break
            print( "retrying to clear %s." % report_folder )

        allow_angular_slope_in_mf = get_setting( 'allow_angular_slope_in_mf' )
        maintenance_mode = get_setting( 'maintenance_mode' )

        clear_settings()

        set_setting('allow_angular_slope_in_mf', allow_angular_slope_in_mf)
        set_setting('maintenance_mode', maintenance_mode)
        set_setting('test_pattern', test_patterns[0])
        set_setting('mapping_image_folder', self.mapping_image_folder)
        set_setting('decomp_image_folder_xray', self.decomp_image_folder_xray)
        set_setting('decomp_image_folder_uv', self.decomp_image_folder_uv)
        set_setting('preview_image_folder', self.preview_image_folder)

        self.report_folder  = report_folder
        self.report_subfolder = report_subfolder
        from molass_legacy.KekLib.TkTester import TestClient
        self.client         = TestClient( self.app, self.test_all_entry, log_func=write_to_log )

    def test_all_entry( self, client, agent ):
        self.client     = client
        self.agent      = agent
        self.counter    = 0
        with_compare = '' if self.to_compare_logfolder is None else ' with compare folder ' + self.to_compare_logfolder
        with_shutdown = ' with shutdown ' if self.shutdown_machine else ''
        write_to_log( 'test start' + with_compare+ with_shutdown + '.\n' )

        for i in self.test_patterns:

            self.test_pattern = i
            # set_setting('test_pattern', self.test_pattern)
            while True:
                if os.path.exists( self.report_folder ):
                    break
                print( "retrying to clear %s." % self.report_folder )
                sleep( 1 )

            if self.test_pattern == 2:
                self.agent.set_setting( 'use_elution_models', 1 )

            agent.analysis_name_entry.focus_force()
            agent.analysis_name_entry.delete( 0, Tk.END)
            if self.counter == 0 and self.report_subfolder is not None and self.report_subfolder !='':
                subfolder = self.report_subfolder
            else:
                subfolder = 'analysis-000'
            agent.analysis_name_entry.insert( 0, subfolder )
            agent.update()

            if self.test_pattern == 33333:
                self.srr_fh = open(self.srr_folder + "/srr.csv", "w")
                if False:
                    self.srr_figs_folder = self.srr_folder + "/figs"
                    if not os.path.exists():
                        os.makedirs(self.srr_figs_folder)

            for k, folder in enumerate(self.test_folders):
                if k == 0:
                    serial_folder_walk(folder, self.test_a_folder)
                elif k == 1:
                    mct_folder_walk(folder, self.test_a_folder)

            if self.test_pattern == 33333:
                self.srr_fh.close()

            if self.test_pattern == -1:
                if self.counter == 0:
                    from importlib import reload
                    import Test.TempInspection
                    reload(Test.TempInspection)
                from molass_legacy.Test.TempInspection import inspection_close
                inspection_close(agent)

            save_folder = self.report_folder + '-' + str(i)
            os.rename( self.report_folder, save_folder )
            clear_dirs_with_retry( [ self.report_folder ] )

        write_to_log( 'done.\n' )
        if self.to_compare_logfolder is not None:
            write_to_log( 'diff  count: %d\n' % self.diff_count )
            write_to_log( 'error count: %d\n' % self.error_count )
            write_to_log( 'mdiff count: %d\n' % self.mdiff_count )
        self.log.close()
        if self.shutdown_machine:
            from molass_legacy.KekLib.Shutdown import shutdown_machine
            shutdown_machine()

    def test_a_folder( self, in_folder, uv_folder, plot=None, suppress_retry_call=False, analysis_name=None ):
        global fromlog_rg_list

        client  = self.client
        agent   = self.agent
        tester_info = Struct(test_pattern=self.test_pattern)
        agent.set_test_mode(tester_info)

        if self.start_str is not None:
            if self.counter == 0:
                self.start_str_found = False
            if in_folder.find( self.start_str ) > 0:
                self.start_str_found = True
            if not self.start_str_found:
                return True, None

        if self.minimum_only:
            self.data_name_key = None
            found = False
            for name_key in MINIMUM_DATA_NAMES:
                if in_folder.find(name_key) >= 0:
                    found = True
                    self.data_name_key = name_key
                    break

            print(in_folder, found)
            if not found:
                return True, None

        if not self.allow_mct_folder:
            from molass_legacy.SerialAnalyzer.SerialDataUtils import get_mtd_filename
            mtd_file = get_mtd_filename(in_folder)
            if mtd_file is not None:
                return True, None

        if self.ds_filter is not None:
            if self.ds_filter.is_in_the_selection(in_folder):
                pass
            else:
                return True, None

        """
        # modify and uncomment to restart
        if self.test_pattern == 5:
            if self.counter == 0:
                self.start_str_found = False
            if in_folder.find( 'Kosugi8' ) > 0:
                self.start_str_found = True
            if not self.start_str_found:
                return True, None
        """

        if self.filter_str is not None and in_folder.find( self.filter_str ) < 0:
            return True, None

        if self.counter > 0:
            sleep( 0.5 )    # to avoid hang?

        self.log.write( 'Doing ' + in_folder + '\n' )
        self.log.flush()

        self.counter += 1

        agent.in_folder_entry.focus_force()
        agent.in_folder_entry.delete( 0, Tk.END)
        agent.in_folder_entry.insert( 0, in_folder )
        agent.on_entry_in_folder()

        if self.test_pattern in [7]:
            agent.fully_automatic.set(1)
        else:
            agent.fully_automatic.set(0)

        if False:
            if uv_folder != in_folder:
                agent.uv_folder_entry.focus_force()
                agent.uv_folder_entry.delete( 0, Tk.END)
                agent.uv_folder_entry.insert( 0, uv_folder )
                agent.on_entry_uv_folder()

        if self.counter == 1:
            agent.an_folder_entry.focus_force()
            agent.an_folder_entry.delete( 0, Tk.END)
            agent.an_folder_entry.insert( 0, self.report_folder)
            agent.on_entry_an_folder()
            agent.update()      # for Windows 7
            sleep( 1 )

        if analysis_name is not None:
            agent.analysis_name.set( analysis_name )

        assert( agent.an_folder.get() == self.report_folder )
        agent.update(__wait__=False )   # for Windows 7

        agent.wait_until_the_data_is_ready()
        if agent.detected_abnomality():
            print( '---- detected_abnomality' )
            try:
                write_from_log_queue()
                write_to_log( '---- detected_abnomality\n' )
                agent.abnomality_dialog.ok_button.invoke()
                sleep( 1 )
                reply_messagebox('Exclusion Done', 'Y')
            except:
                # this occurs in safe mode
                print('assumed to have been replied manually.')
            sleep( 1 )

        if self.test_pattern == -1:
            if self.counter == 0:
                from importlib import reload
                import Test.TempInspection
                reload(Test.TempInspection)
            from molass_legacy.Test.TempInspection import datarange_problem_inspection
            datarange_problem_inspection(agent, self.result_folder)
            if self.debug_trial and self.counter > 2:
                return False, None
            else:
                return True, None

        agent.set_setting('suppress_low_quality_warning', 1)
        agent.set_test_mode( Struct( test_pattern=self.test_pattern ) )

        if self.test_pattern >= 20 and self.test_pattern < 30:
            self.do_v2_results(in_folder)
            return True, None

        if self.test_pattern in [0]:
            self.do_outline_only()
            return True, None

        if self.test_pattern >= 30:
            self.do_v2_thru_peak_editor()
            return True, None

        if self.test_pattern == 8:
            try:
                self.do_v2_demo()
                return True, None
            except:
                return False, None

        analyzer    = agent.analyzer

        if self.save_outline_figures:
            agent.fig_frame.save_the_figure( self.outline_image_folder, agent.get_analysis_name() )

        if self.test_pattern in [7]:
            return self.do_fully_automatic()

        if self.test_pattern in [5]:
            if self.minimum_only:
                path_length = agent.get_setting_for_test('path_length')
                assert path_length == PATH_LENGTH_DICT[self.data_name_key]

            agent.refresh_button.invoke()
            agent.wait_until_the_data_is_ready()

        if self.test_pattern in [4, 5, 6]:

            if not agent.doing_mfc():

                if self.save_3d_figures:
                    agent.presync_button.invoke(__wait__=False )
                    sleep(1)
                    threedim_dialog = agent.threedim_dialog
                    threedim_dialog.save_the_figure( self.threed_image_folder, agent.analysis_name.get() )
                    threedim_dialog.cancel()

                if self.test_pattern in [5]:
                    agent.restrict_button.invoke(__wait__=False)

                    sleep(1)
                    while True:
                        datarange_dialog = agent.datarange_dialog
                        if datarange_dialog is not None:
                            break
                        print("waiting for datarange_dialog being ready")
                        sleep(1)

                    if not agent.get_setting_for_test('use_xray_conc'):

                        for i in range(3):
                            sleep(1)
                            try:            
                                current_frame = datarange_dialog.get_current_frame()
                                current_frame.toggle_btn.invoke()   # note that current_frame is not an agent widget
                                sleep(1)
                                current_frame = datarange_dialog.get_current_frame()
                                current_frame.toggle_btn.invoke()   # note that current_frame is not an agent widget
                                break
                            except Exception as e:
                                print(e)
                                print([i], "retrying...")
                                continue

                        sleep(1)

                    # datarange_dialog.save_the_figure( self.restrictor_image_folder, agent.analysis_name.get() )
                    datarange_dialog.cancel()

                if self.save_baseline_figures:
                    agent.baseline_button.invoke(__wait__=False )
                    sleep(3)
                    baseline_inspector = agent.baseline_inspector
                    baseline_inspector.save_the_figure( self.baseline_image_folder, agent.analysis_name.get() )
                    baseline_inspector.cancel()
                    agent.update()

            sleep(1)
            # return True, None

        agent.analysis_button.invoke( __wait__=False )
        sleep( 1 )      # take some time to get analyzer.mapper_canvas cleared

        if agent.doing_mfc():
            self.do_mfc()
            return True, None

        while not analyzer.has_mapper_canvas():
            sleep( 1 )

        mapper_canvas = analyzer.mapper_canvas
        if not mapper_canvas.get_edit_mode():
            mapper_canvas.toggle_button.invoke()
        sleep( 1 )

        if self.test_pattern in [33333]:
            mapper_canvas.cancel_button.invoke()
            sleep(0.5)
            scd_list = mapper_canvas.get_cd_list()
            sd = mapper_canvas.get_sd()
            from Rank.SRR import SRR
            srr = SRR(sd)
            # srr_list = srr.compute_judge_info(figs_folder=self.srr_figs_folder, analysis_name=agent.analysis_name.get())
            srr_list = srr.compute_judge_info()
            for k, (scd, rec) in enumerate(zip(scd_list, srr_list)):
                self.srr_fh.write(','.join([in_folder] + ["%s" % v if v is None else "%g" % v for v in (k, scd, *rec)]) + "\n")
                self.srr_fh.flush()
            return True, None

        if self.test_pattern in [4]:
            mapper_canvas.mapper.make_uniformly_scaled_vector(scale=1, survey_fh=self.survey_fh)
            if self.save_mapping_figures:
                mapper_canvas.plotter.save_the_figure( self.mapping_image_folder, agent.analysis_name.get() )
            if True:
                mapper_canvas.cancel_button.invoke()
                sleep(0.5)
                return True, None
            else:
                sleep(0.5)

        if self.test_pattern in [9, 10]:
            try:
                self.do_fullopt_initialization(mapper_canvas)
                if self.test_pattern == 10:
                    self.do_fullopt_execution(mapper_canvas)

                if self.debug_trial:
                    ret_tuple = False, None
                else:
                    ret_tuple = True, None
            except:
                ret_tuple = False, None
            if self.save_mapping_figures:
                mapper_canvas.plotter.save_the_figure( self.mapping_image_folder, agent.analysis_name.get() )
            mapper_canvas.cancel_button.invoke()
            sleep(0.5)
            return ret_tuple

        adjuster = mapper_canvas.adjuster
        conc_type = adjuster.get_conc_type()

        # at this time, app_logfile should have already been set
        self.app_logfile = analyzer.get_logfile()
        if self.to_compare_logfolder is None:
            self.from_logfile   = None
        else:
            self.from_logfile   = '/'.join(  [self.to_compare_logfolder] + self.app_logfile.split( '/' )[-2:] )

        if INVESTIGATE_ABSORBANCE_ANOMALY:
            anomaly = adjuster.observe_base_curve_anomaly()
            print( 'anomaly=', str(anomaly) )

        if self.test_pattern in [1, 2]:
            adjuster.advanced_btn.invoke( __wait__=False )
            sleep( 0.5 )
            dialog = adjuster.advanced_settings_dialog
            dialog.range_type_buttons[1].invoke()
            sleep( 1 )
            dialog.ok_button.invoke()
            sleep( 0.5 )

        if self.test_pattern in [6] and conc_type < 2:
            if self.save_cdi_figures:
                plotter = mapper_canvas.plotter
                for pno in range(len(plotter.get_target_ranges())):
                    plotter.show_cdi_dialog(pno=pno, show_later=True)
                    plotter.cdi_dialog.show( __wait__=False )
                    sleep(1)
                    plotter.cdi_dialog.save_the_figure(self.cdi_image_folder, pno)
                    plotter.cdi_dialog.ok()
                    sleep(0.5)

        if self.test_pattern in [4, 5, 6] and conc_type < 2:
            if self.test_pattern in [4, 6]:
                need_decomp = mapper_canvas.guide_info.decomp_proc_needed()
            else:
                need_decomp = True

            def preview_control(need_decomp, post_fix=""):
                if need_decomp:
                    mapper_canvas.decomp_btn.invoke( __wait__=False )
                    sleep( 1 )
                    decomp_editor = mapper_canvas.decomp_editor
                    while not decomp_editor.is_ready():
                        print( 'waiting for the decomp_editor to be ready.' )
                        sleep( 1 )

                    if self.save_decomp_figures:
                        decomp_editor.save_the_figure( self.decomp_image_folder_xray, agent.analysis_name.get() )

                    if self.test_pattern != 4:

                        decomp_editor.get_current_frame().toggle_btn.invoke()

                        if self.save_decomp_figures:
                            decomp_editor.save_the_figure( self.decomp_image_folder_uv, agent.analysis_name.get() )

                    write_to_log( 'model_fit_error=' + str(mapper_canvas.decomp_editor.get_model_fit_errors()) + '\n' )
                    write_to_log( 'applied model=' + decomp_editor.get_applied_model().get_name() + '\n' )

                    if self.test_pattern != 4:
                        self.do_preview(decomp_editor, post_fix)

                    decomp_editor.ok_button.invoke()
                    print( 'decomp_editor ok' )
                    sleep( 0.5 )
                else:
                    mapper_canvas.reditor_btn.invoke( __wait__=False )
                    sleep( 1 )
                    range_editor = mapper_canvas.range_editor
                    while not range_editor.is_ready():
                        print( 'waiting for the range_editor to be ready.' )
                        sleep( 1 )

                    if self.test_pattern != 4:
                        self.do_preview(range_editor, post_fix)
                    range_editor.ok_button.invoke()
                    print( 'range_editor ok' )
                    sleep( 0.5 )

            if self.mapping_only:
                pass
            else:
                if self.test_pattern == 44444:
                    conc_type = adjuster.get_conc_type()
                    if conc_type == 0:
                        adjuster.uv_baseline_adjust.set(0)
                        adjuster.optimize_btn.invoke()
                        preview_control(need_decomp, post_fix="-bpa")
                        for bpa_var in [adjuster.uv_baseline_with_bpa, adjuster.xray_baseline_with_bpa]:
                            bpa_var.set(0)
                        adjuster.optimize_btn.invoke()
                        preview_control(need_decomp, post_fix="-nonbpa")
                else:
                    preview_control(need_decomp)

        num_peaks = mapper_canvas.mapper.get_num_peaks()
        write_to_log( 'num_peaks=' + str(num_peaks) + '\n' )

        if self.test_pattern in [0, 2, 5, 6]:
            mapper = mapper_canvas.mapper
            sci = mapper.get_sci_list()
            write_to_log( 'sci=' + str(sci) + '\n' )

            mapped_info = mapper.get_mapped_info()
            put_to_tester_log_queue( 'x_ranges=%s\n' % str(mapped_info.x_ranges) )

        if self.save_mapping_figures:
            mapper_canvas.plotter.save_the_figure( self.mapping_image_folder, agent.analysis_name.get() )

        if self.mapping_only:
            mapper_canvas.cancel_button.invoke()
            sleep(0.5)
        else:
            mapper_canvas.ok_button.invoke()
            sleep(0.5)
            self.do_whole()

        self.do_finish( in_folder )

        # if self.counter > 3: return False, None
        # TODO: return serial_data
        if self.debug_trial:
            # stop on return
            return False, None
        else:
            return True, None

    def do_preview(self, dialog, post_fix):
        preview_frame = dialog.preview_frame
        preview_frame.show_zx_preview_button.invoke( __wait__=False )
        sleep( 1 )
        pool = preview_frame.pool
        preview_wait_count = 0
        while not pool.ok() and preview_wait_count < 1200:
            print( 'preview_wait_count=', preview_wait_count )
            sleep( 0.5 )
            preview_wait_count += 1
        if pool.ok():
            if self.save_preview_figures:
                sleep( 0.5 )
                pool.dialog.save_the_figure( self.preview_image_folder, self.agent.analysis_name.get() )
                solver_results, peak_range_infos = pool.get_solver_ret_tuple()
            if self.save_preview_results:
                pool.dialog.save_results( __wait__=False )
                sleep(0.5)
                saver = pool.dialog.saver
                ret = saver.validate(notify=False, post_fix=post_fix)
                if ret:
                    print('preview results save ok.')
                else:
                    print('preview results save failed.')
                saver.cancel()
            if self.save_preview_cdi_figures:
                pool.dialog.auto_cdi_for_all(out_folder=self.preview_cdi_image_folder)
            pool.dialog.ok()
        else:
            put_to_tester_log_queue( 'failed to preview' )

    def do_whole( self ):
        global fromlog_rg_list

        agent           = self.agent
        from_logfile    = self.from_logfile
        analyzer        = agent.analyzer
 
        if self.to_compare_logfolder is None:
            fromlog_rg_list = None
        else:
            # print( 'from_logfile=', from_logfile )
            if os.path.exists( from_logfile ):
                fromlog_rg_list = []
                final_std_diff  = None
                try:
                    fh = open( from_logfile )
                    for line in fh:
                        if final_std_diff is None:
                            p = line.find( final_std_diff_str )
                            if p > 0:
                                p_ = p + len(final_std_diff_str)
                                # print( 'line[p_:]=-1', line[p_:-1] )
                                final_std_diff = line[p_:-1]
                                self.log_final_std_diff_check( final_std_diff )
                                continue

                        if line.find( 'Rg' ) < 0:
                            continue
                        m = success_rg_re.search( line )
                        if m:
                            fromlog_rg_list.append( [ float( m.group(i) ) for i in [2, 3] ] )
                            continue
                        m = failure_rg_re.search( line )
                        if m:
                            fromlog_rg_list.append( None )
                            continue
                    # print( 'fromlog_rg_list=', fromlog_rg_list )
                except:
                    from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                    etb = ExceptionTracebacker()
                    self.log.write( str( etb ) )
                    fromlog_rg_list = None
                # self.log.write( str( fromlog_rg_list ) + '\n' )
            else:
                fromlog_rg_list = None

        write_from_log_queue()

        analyzer.dialog.run_button.invoke( __wait__=False )

        counter = 0
        exception_break = False
        while True:
            counter += 1
            # if analyzer.progress_dialog.is_at_end(): # seems sometimes to fail on Windows 7
            try:
                if analyzer.progress_dialog.button.cget( 'text' ) == "OK":
                    break
            except:
                etb = ExceptionTracebacker()
                write_to_log( str(etb) )
                print( 'Canceled?' )
                exception_break = True
                break

            if not agent.__is_alive__():
                write_from_log_queue()
                write_to_log( 'agent is not alive\n' )
                return False, None

            sleep( 1 )
            print( 'counter=', counter )

        write_from_log_queue()

        if not exception_break:
            analyzer.progress_dialog.button.invoke()
        agent.update()
        sleep( 1 )

    def do_fully_automatic( self ):
        agent = self.agent
        if self.test_pattern == 6:
            agent.analyzer.set_cancel_mode()

        # agent.fully_automatic_cb.invoke() is not appropriate here because it toggles at the next time
        agent.fully_automatic.set(1)
        agent.analysis_button.invoke( __wait__=False )
        sleep(1)
        reply_messagebox('Fully Automatic Run Confirmation', 'Y')
        sleep(1)
        log_confirm = 'Log file removal confirmation'
        while agent.is_busy():
            print('waiting for the fully_automatic control to finish.')
            sleep(1)
            if self.test_pattern == 6:
                print('window_exists(log_confirm)')
                if window_exists(log_confirm):
                    sleep(1)
                    print('reply_messagebox')
                    reply_messagebox(log_confirm, 'N')
                    print('reply_messagebox end')
                print('test_pattern 6 continue')
        return True, None

    def do_mfc(self):
        write_to_log( 'skippig MFC data.\n' )
        while True:
            try:
                self.agent.analyzer.mfc_dialog.cancel()
                break
            except:
                sleep(1)

    def do_finish( self, in_folder ):
        app_logfile     = self.app_logfile
        from_logfile    = self.from_logfile

        write_from_log_queue()

        if self.to_compare_logfolder is not None:
            try:
                diff_count, error_count = self.make_log_diff( app_logfile, from_logfile )
                print( 'diff_count, error_count=', diff_count, error_count  )
                self.diff_count += diff_count
                self.error_count += error_count
            except:
                etb = ExceptionTracebacker()
                print( etb )
            sleep( 1 )

        self.log.write( 'Finished ' + in_folder + '\n' )

    def log_final_std_diff_check( self, final_std_diff ):
        if final_std_diff == self.final_std_diff:
            write_to_log( 'final_std_diff ok\n' )
        else:
            write_to_log( 'final_std_diff is different: %s != %s\n' % ( final_std_diff, self.final_std_diff ) )
            self.mdiff_count += 1

    def get_q_statistics( self ):
        sd = self.app.analyzer.dialog.serial_data
        qvector = sd.qvector
        qmin = np.min(qvector)
        qmax = np.max(qvector)
        qdelta = qvector[1] - qvector[0]
        size = len(qvector)
        imax = np.max( sd.intensity_array[:, :, 1] )
        return [qmin, qmax, qdelta, size, imax]

    def make_log_diff( self, app_logfile, from_logfile ):
        from molass_legacy.KekLib.LogDiff import make_log_diff, time_re_sub
        to_logfile   = app_logfile

        def my_re_sub( line ):
            line = time_re_sub( line )
            line = re.sub( success_rg_re, 'Rg from (ffff.dat) is estimated to be (00.0) with quality (0.00).', line )
            line = re.sub( pytools_ver_rg, 'pytools-x_x_x-develop', line )
            return line

        diff_count = 0
        error_count = 0
        try:
            diff = make_log_diff( from_logfile, to_logfile, re_sub=my_re_sub )
            if len(diff) > 0:
                self.log.write( '==== diff begin ====\n' )
                self.log.write( diff )
                self.log.write( '==== diff end ======\n' )
                diff_count = 1
            else:
                self.log.write( '==== no diff in the log file ====\n' )
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.log.write( 'make_log_diff failed: ' + str(etb) + '\n' )
            error_count = 0

        return diff_count, error_count

    def do_outline_only(self):
        folder, _ = os.path.split(self.report_folder)

        if self.outline_image_folder is None:
            self.outline_image_folder = os.path.join(folder, 'images-outline').replace( '\\', '/' )
            clear_dirs_with_retry([self.outline_image_folder])

        agent = self.agent
        if agent.doing_mfc():
            put_to_tester_log_queue('Skipped MFC data')
            return

        agent.get_it_ready_for_qmm()
        agent.fig_frame.save_the_figure( self.outline_image_folder, agent.analysis_name.get() )
        return

    def do_v2_demo(self):
        folder, _ = os.path.split(self.report_folder)

        if self.qmm_image_folder is None:
            self.qmm_image_folder = os.path.join(folder, 'qmm-images').replace( '\\', '/' )
            clear_dirs_with_retry([self.qmm_image_folder])

        agent = self.agent
        if agent.doing_mfc():
            put_to_tester_log_queue('Skipped MFC data')
            return

        if True:    # add another arg to select between EGH and EMG
            agent.qmm_menu.show_egh_qmm_dialog(__wait__=False)
        else:
            agent.qmm_menu.show_emg_qmm_dialog(__wait__=False)

        # must do this after agent.get_it_ready_for_qmm() in the above call
        if self.outline_image_folder is not None:
            agent.fig_frame.save_the_figure( self.outline_image_folder, agent.analysis_name.get() )

        while agent.is_qmm_busy():
            print("waining for the qmm-controller's getting ready.")
            sleep(1)

        qmm_controller = agent.get_qmm_controller()
        if qmm_controller is None:
            put_to_tester_log_queue('Failed to get qmm_controller')
            assert False

        try:
            while not qmm_controller.dialog_ready():
                print("waining for the qmm-dialo's getting ready.")
                sleep(1)
        except:
            put_to_tester_log_queue('Failed in dialog_ready')
            assert False

        sleep(1)
        qmm_dialog = qmm_controller.dialog
        frame = qmm_dialog.get_current_frame()
        path = frame.save_the_figure(self.qmm_image_folder, agent.analysis_name.get())
        qmm_dialog.cancel()
        agent.clear_test_mode()
        write_from_log_queue()

        file = path + '.png'
        if not os.path.exists(file):
            put_to_tester_log_queue('Failed to save the figure' + file)
            write_from_log_queue()
            agent.quit(immediately=True)
            assert False

    def do_fullopt_initialization(self, mapper_canvas):
        print("start do_fullopt_initialization")
        mapper_canvas.full_opt_button.invoke(__wait__=False)
        sleep(3)

        while True:
            try:
                fullopt_init = mapper_canvas.fullopt_init
                while fullopt_init.is_busy():
                    print('waiting for the fullopt_init to be ready.')
                    sleep(1)

                fullopt_init.save_the_figure(self.fulloptinit_image_folder, self.agent.analysis_name.get())
                break
            except:
                print("May be we can't get fullopt_init")
                sleep(1)
                continue

        if self.test_pattern == 9:
            fullopt_init.user_cancel(ask=False)
        sleep(1)

    def do_fullopt_execution(self, mapper_canvas):
        print("start do_fullopt_execution")
        fullopt_init = mapper_canvas.fullopt_init
        fullopt_init.proceed_btn.invoke(__wait__=False)     # fullopt_init.ok(__wait__=False) makes it freeze
        sleep(1)

        while True:
            try:
                fullopt_dialog = mapper_canvas.fullopt_dialog
                while fullopt_dialog.is_busy():
                    t = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    print('%s waiting for the fullopt_dialog to be ready.' % t)
                    progress = fullopt_dialog.get_progress()
                    seconds = 60 if progress < 0.98 else 10
                    sleep(seconds)
                break
            except:
                print("May be we can't get fullopt_dialog")
                sleep(1)
                continue

        fullopt_dialog.cancel()
        sleep(1)

    def make_v2_results_dict(self):
        analysis_folder_re = re.compile("(analysis-\d+)")
        in_folder_re = re.compile("in_folder=(\S+)")
        self.results_dict = {}
        for folder in self.v2_analysis_folders:
            in_data_info_txt = os.path.join(folder, r"optimizer\000\in_data_info.txt")
            if os.path.exists(in_data_info_txt):
                with open(in_data_info_txt) as fh:
                    line = fh.read()
                m = in_folder_re.match(line)
                assert m
                in_folder = m.group(1)

                job_folder = os.path.join(folder, r"optimizer\000")
                print(in_folder, job_folder)
                self.results_dict[in_folder] = job_folder

    def do_v2_results(self, in_folder):
        job_folder = self.results_dict.get(in_folder)
        if job_folder is None:
            print("-------------------- No results exist for %s" % in_folder)
            return

        agent   = self.agent
        agent.v2_menu.show_result_trace(__wait__=False)
        sleep(1)
        trace_dialog = agent.v2_menu.get_trace_dialog()

        entry = trace_dialog.optjob_folder_entry
        entry.delete(0, Tk.END)
        entry.insert(0, job_folder)

        callback_txt = os.path.join(job_folder, "callback.txt")
        size = os.path.getsize(callback_txt)
        if size == 0:
            # as with 20160628
            print("It seems like a failed result. Skip.")
            trace_dialog.cancel()
            return

        try:
            trace_dialog.on_entry_optjob_folder()
            trace_dialog.proceed_btn.invoke(__wait__=False)

            i = 0
            while True:
                print([i], "waiting for FullOptDialog's being ready.")
                sleep(1)
                fullopt_dialog = agent.get_fullopt_dialog()
                if fullopt_dialog is not None:
                    break
                i += 1

            print([i], "FullOptDialog seems to be ready.")
            fullopt_dialog.runner.set_work_folder(job_folder)
            fullopt_dialog.save_the_result_figure()
            fullopt_dialog.cancel_button.invoke()   # fullopt_dialog.cancel() gets an error
        except Exception as e:
            print(e)

        sleep(1)

    def do_v2_thru_peak_editor(self):
        agent   = self.agent

        agent.v2_menu.show_peak_editor(__wait__=False)
        i = 0

        while True:
            print([i], "waiting for OptStrategyDialog's construction.")
            sleep(1)
            strategy_dialog = agent.v2_menu.get_strategy_dialog()
            if strategy_dialog is not None:
                break
            i += 1

        strategy_dialog.proceed_btn.invoke()

        while True:
            print([i], "waiting for PeakEditor's construction.")
            sleep(1)
            peak_editor = agent.v2_menu.get_peak_editor()
            if peak_editor is not None:
                break
            i += 1

        while peak_editor.is_busy():
            print([i], "waiting for PeakEditor's being ready.")
            sleep(1)
            i += 1

        print([i], "PeakEditor seems to be ready.")
        if self.save_peakeditor_figures:
            analysis_name = agent.analysis_name.get()
            filename = analysis_name.replace( 'analysis', 'figure' )
            file_path = os.path.join(self.peakeditor_image_folder, filename)
            peak_editor.save_the_figure(file_path)

        if self.test_pattern == 30:
            peak_editor.cancel_btn.invoke()      # peak_editor.user_cancel(ask=False) gets "RuntimeError: main thread is not in main loop"
        else:
            peak_editor.proceed_btn.invoke()
            while True:
                print([i], "waiting for FulloptDialog's construction.")
                sleep(1)
                fullopt_dialog = agent.v2_menu.get_fullopt_dialog()
                if fullopt_dialog is not None:
                    break
                i += 1
            self.wait_for_v2_completion(fullopt_dialog)

    def wait_for_v2_completion(self, fullopt_dialog):
        i = 0
        while fullopt_dialog.is_busy():
            print([i], "waiting for FulloptDialog's being complete.")
            sleep(60)
            i += 1

        fullopt_dialog.cancel_button.invoke()



"""
    ModelDebugUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt

def model_plot(x, y, xr_cy_list, xr_ty, uv_x, uv_y, uv_cy_list, uv_ty):

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))

        ax1.plot(uv_x, uv_y)
        for k, cy in enumerate(uv_cy_list):
            ax1.plot(uv_x, cy, ":", label="c=%d"%k)
        ax1.plot(uv_x, uv_ty, ":", color="red", label="c-total")
        ax1.legend()

        ax2.plot(x, y)
        for k, cy in enumerate(xr_cy_list):
            ax2.plot(x, cy, ":", label="c=%d"%k)
        ax2.plot(x, xr_ty, ":", color="red", label="c-total")
        ax2.legend()

        fig.tight_layout()
        plt.show()



"""
    Optimizer.LrfOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize

def optimize_decomposition_for_lrf(optimizer, params):

    # indeces = optimizer.get_lrfopt_indeces()
    indeces = np.array([], dtype=int)

    temp_params = params.copy()

    def non_negative_lrf_objective(p):
        temp_params[indeces] = p
        fv = optimizer.objective_func(temp_params)


"""
    OptJobResultInfo.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

class OptJobResultInfo:
    def __init__(self, fv=None, params=None):
        self.fv = params
        self.params = params



"""
    Theory.BoundedLRF.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize
from matplotlib.patches import Rectangle
from time import time
from molass_legacy.KekLib.SciPyCookbook import smooth
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data
from .SolidSphere import phi
from .Rg import compute_corrected_Rg
from .SphericalFit import estimate_LR
from .SolidSphere import get_boundary_params_simple
from Rank.Synthesized import synthesized_data

def fit_exp_dist(p, py):
    apy = np.abs(py)

    def obj_func(x):
        K, L, M = x
        v = K*L*np.exp(-L*M*p)
        return np.sum((v - apy)**2)

    res = minimize(obj_func, (1,1,1))

    return res.x

def spike():

    qv = np.linspace(0.005, 0.5, 100)
    R = np.sqrt(5/3)*30
    K = 5

    n = int(R*K*qv[-1]/(2*np.pi))
    print('n=', n)
    pp = np.arange(1, n)*2*np.pi/(R*K)
    pn = (np.arange(1, n)*2-1)*np.pi/(R*K)

    Aq = phi(qv, R)**2
    Sf_1 = -phi(qv, K*R)
    Bq = Aq*Sf_1

    ppy = -phi(pp, K*R)
    pny = -phi(pn, K*R)

    show_fn_bound = False

    if show_fn_bound:
        K, L, M = fit_exp_dist(pn, pny)
        print('K,L,M=', K, L, M)
        ey = K*L*np.exp(-L*M*qv)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))

    for i, ax in enumerate([ax1, ax2]):

        ax.set_title("B(q)/A(q) Upper and Lower Bounds" + ("" if i==0 else " (Zoomed)"), fontsize=16)
        ax.set_yscale('log')

        if i == 0:
            ax.plot(qv, Aq)
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
        else:
            ax.set_xlim(xmin, xmin*(1-0.25) + xmax*0.25)
            ax.set_ylim(np.exp(np.log10(ymin)*(1-0.75) + np.log10(ymax)*0.75), ymax)
            # ax.set_axis_off()

        axt = ax.twinx()
        axt.grid(False)
        axt.plot(qv, Sf_1, color='yellow', label='B(q)/A(q)')
        axt.plot(qv, Bq, color='cyan', label='B(q)')

        axt.plot(pp, ppy, 'o:', color='red', label='upper bound')
        axt.plot(pn, pny, 'o:', color='blue', label='lower bound')

        if show_fn_bound:
            for j, y in enumerate([ey, -ey]):
                axt.plot(qv, y, ':', color='green', label='functional bound' if j==0 else None)

        ymint, ymaxt = axt.get_ylim()
        axt.set_ylim(ymint, ymaxt)

        b = pn[1]
        axt.plot([b, b], [ymint, ymaxt], ':', color='gray', label='rank boundary')

        axt.legend(loc='lower center' if i==0 else 'lower right')

    fig.tight_layout()
    plt.show()

def plot_P(ax, qv, pt, P):
    aq = P[:,0]
    bq = P[:,1]

    aq_ = aq/np.max(aq)

    ax.set_yscale('log')
    axt = ax.twinx()

    pt_ = pt/np.max(pt)

    ax.plot(qv, pt_, label='peak top')
    ax.plot(qv, aq_, label='A(q)')
    axt.plot(qv, bq, label='B(q)', color='pink')

    ax.legend(fontsize=16)
    axt.legend(bbox_to_anchor=(1, 0.82), loc='upper right')

MINBOUND = 1e-8

def bounds(qv, aq, starv, px, py):

    def minbound(v):
        return max(MINBOUND, v) if v > 0 else min(-MINBOUND, v)

    i = 0
    j = 0
    bv = np.zeros(len(qv))
    for k, x in enumerate(px):
        while qv[j] < x:
            j +=1

        # print([k], x, (i, j))
        if i == 0:
            bv[i:j] = starv
        else:
            n = j - i
            w = np.arange(n)/n
            bv[i:j] = minbound(aq[i] * py[k-1]) * (1-w) + minbound(aq[j] * py[k]) * w

        i = j
    if j < len(bv):
        bv[j:] = minbound(aq[i] * py[k])

    return bv

def construct_bounds(qv, R, P, rscale, bscale, debug=False):
    n = int(R*rscale*qv[-1]/(2*np.pi))
    print('n=', n)

    pp = np.arange(1, n)*2*np.pi/(R*rscale)
    pn = (np.arange(1, n)*2-1)*np.pi/(R*rscale)

    ppy = -phi(pp, rscale*R)
    pny = -phi(pn, rscale*R)

    print(np.min(ppy), np.max(pny))

    # aq = P[:,0]
    aq = phi(qv, R)**2
    lowerv = bounds(qv, aq, -np.inf, pn, pny)
    upperv = bounds(qv, aq, +np.inf, pp, ppy)

    if debug:
        Aq = phi(qv, R)**2
        Sf_1 = -phi(qv, rscale*R)
        aq = P[:,0]
        bq = P[:,1]
        minaq = np.percentile(aq, 10)
        aq_ = np.max([aq, np.ones(len(aq))*minaq], axis=0)
        sf_1 = bscale*bq/aq_

        print('bscale=', bscale)
        plt.push()
        fig, axes = plt.subplots(ncols=2, figsize=(16,7))

        for i, ax in enumerate(axes):
            ax.plot(qv, bq, color='pink')
            ax.plot(qv, Sf_1, ':', color='green')
            # ax.plot(qv, sf_1, color='yellow')
            if i == 0:
                ax.plot(pp, ppy, 'o:', color='red', label='upper bound')
                ax.plot(pn, pny, 'o:', color='blue', label='lower bound')
            else:
                ax.plot(qv, upperv, ':', color='red', label='upper bound')
                ax.plot(qv, lowerv, ':', color='blue', label='lower bound')

        fig.tight_layout()
        plt.show()
        plt.pop()

    return lowerv, upperv

def optimize_impl(P, C, M, lbound, ubound, boundary_info=None):

    # method_opts = ['Nelder-Mead', False]  # it takes too long
    # method_opts = ['Powell',      True]   # it takes too long
    method_opts = ['CG', False]             #
    # method_opts = ['BFGS', False]         # Method BFGS cannot handle constraints nor bounds.

    method, bounds_ok = method_opts

    print('method=', method, bounds_ok)

    t0 = time()

    xinit = P.flatten()
    isize = P.shape[0]

    if bounds_ok:
        def obj_func(x):
            P_ = x.reshape(P.shape)
            return np.linalg.norm(P_@C - M)
        lbound_ = np.hstack([-np.ones(isize)*np.inf, lbound])
        ubound_ = np.hstack([+np.ones(isize)*np.inf, ubound])
        res = minimize(obj_func, xinit, bounds=np.vstack([lbound_, ubound_]).T, method=method)

    else:
        assert boundary_info is not None

        if False:
            qv, k, b = boundary_info

            boundary_i = bisect_right(qv, b)
            print('b=', b, 'boundary_i=', boundary_i)
            bslice = slice(boundary_i, None)
            zeros = np.zeros(isize-boundary_i)
            lbound_ = lbound[bslice]
            ubound_ = ubound[bslice]

            def obj_func_regular1(x):
                P_ = x.reshape(P.shape)
                norm = np.linalg.norm(P_@C - M)
                bq = P_[bslice,1]
                penalty = ( np.sum(np.min([bq - lbound_, zeros], axis=0)**2)
                          + np.sum(np.max([bq - ubound_, zeros], axis=0)**2)
                          )
                return norm + penalty*10

            res = minimize(obj_func_regular1, xinit, method=method)
        else:
            qv, k, b = boundary_info

            zeros = np.zeros(isize)
            penalty_weights_logistic = 1/(1+np.exp(-k*(qv-b)))
            if False:
                denom = 1/np.max([np.abs(lbound), np.abs(ubound)], axis=0)
                normal_denom = denom/np.sum(denom)
                penalty_weights_total = penalty_weights_logistic*normal_denom
            penalty_weights_total = penalty_weights_logistic

            if True:
                plt.push()
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))
                ax1.plot(qv, lbound)
                ax1.plot(qv, ubound)
                ax2.plot(qv, penalty_weights_logistic)
                # ax2.plot(qv, normal_denom)
                # ax2.plot(qv, penalty_weights_total)
                fig.tight_layout()
                plt.show()
                plt.pop()

            def obj_func_regular2(x):
                P_ = x.reshape(P.shape)
                norm = np.linalg.norm(P_@C - M)
                bq = P_[:,1]
                penalty = ( np.sum(np.min([bq - lbound, zeros]*penalty_weights_total, axis=0)**2)
                          + np.sum(np.max([bq - ubound, zeros]*penalty_weights_total, axis=0)**2)
                          )
                return norm + penalty*100

            res = minimize(obj_func_regular2, xinit, method=method)

    Popt = res.x.reshape(P.shape)

    print('res.success=', res.success)
    print('res.status=', res.status)
    print('res.nit=', res.nit)

    print("It took", time()-t0)

    return Popt

def demo(root, sd, pno=0, synthesized_lrf=None, bounded_lrf=True, debug=True):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    pt = M[:,p]

    x = ecurve.x
    y = ecurve.y

    i = bisect_right(qv, 0.02)

    eslice = slice(f,t)
    M0 = M[:,eslice]
    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])

    norm_list = []

    c_ = y[eslice]

    M2 = get_denoised_data(M0, rank=2)
    C2 = np.array([c_, c_**2])
    P2 = M2 @ np.linalg.pinv(C2)

    norm_list.append(np.linalg.norm(P2@C2 - M0))

    b1, b2, k = get_boundary_params_simple(Rg)
    L, _, rscale, bscale, score = estimate_LR(qv, M[:,eslice], Rg, b1, debug=False)

    M_, _ = synthesized_data(qv, M[:,eslice], Rg, rank=2, cd=2, boundary=b1, k=k)
    P_ = M_ @ np.linalg.pinv(C2)

    norm_list.append(np.linalg.norm(P_@C2 - M0))

    if bounded_lrf:
        R = np.sqrt(5/3)*Rg
        lbound, ubound = construct_bounds(qv, R, P_, rscale, bscale, debug=True)

        Popt_ = optimize_impl(P_, C2, M2, lbound, ubound, boundary_info=(qv, k, b1))
        norm_list.append(np.linalg.norm(Popt_@C2 - M0))
    else:
        Popt_ = None

    if synthesized_lrf is None:
        Popt = Popt_

        ax3_tilte = "Bounded LRF"
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21,11))

        ax1, ax2, ax3 = axes[0,:]
        ax4, ax5, ax6 = axes[1,:]
    else:
        Popt = synthesized_lrf(qv, M[:,eslice], c_, M2, P2, boundary=b1, k=k)
        norm_list.append(np.linalg.norm(Popt@C2 - M0))

        ax3_tilte = "Result Synthesized LRF"
        fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(22,10))

        ax1, ax2, ax3 = axes[0,0:3]
        ax4, ax5, ax6 = axes[1,0:3]
        axu, axd = axes[:,3]
        axu.set_title("Bounded LRF", fontsize=20)

    ax1.set_title("Rank(2,2) LRF", fontsize=20)
    ax2.set_title("Data Synthesized LRF", fontsize=20)
    ax3.set_title(ax3_tilte, fontsize=20)

    if False:
        ax1.plot(x, y, color='orange')
        ymin, ymax = ax1.get_ylim()
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax1.add_patch(p)

    plot_P(ax1, qv, pt, P2)
    plot_P(ax2, qv, pt, P_)
    plot_P(ax3, qv, pt, Popt)
    axes_ = [ax2, ax3]
    if bounded_lrf:
        plot_P(axu, qv, pt, Popt_)
        axes_.append(axu)

    for ax in axes_:
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        ax.plot([b1, b1], [ymin, ymax], ':', label='rank boundary', color='gray')
        ax.legend(fontsize=16)

    ax4.plot(qv, P2[:,1]/P2[:,0], label='B(q)/A(q)')
    ax5.plot(qv, P_[:,1]/P_[:,0], label='B(q)/A(q)')
    ax6.plot(qv, Popt[:,1]/Popt[:,0], label='B(q)/A(q)')
    axes_ = [ax4, ax5, ax6]
    if bounded_lrf:
        axd.plot(qv, Popt_[:,1]/Popt_[:,0], label='B(q)/A(q)')
        axes_.append(axd) 

    for ax in axes_:
        ax.set_ylim(-2, 2)
        xmin, xmax = ax.get_xlim()
        ax.set_xlim(xmin, xmax)
        ax.plot([xmin, xmax], [0, 0], ':', color='red')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        ax.plot([b1, b1], [ymin, ymax], ':', label='rank boundary', color='gray')
        ax.legend(loc='upper left', fontsize=16)

    fig.tight_layout()

    print('norm_list=', norm_list)
    plt.show()



"""
    ElutionModelScaler.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from scipy.optimize import minimize
from molass_legacy.KekLib.BasicUtils import Struct, ordinal_str
import molass_legacy.KekLib.DebugPlot as plt

REMOVE_H_RATIO_LIMIT = 0.01     # < 0.0199 for HasA(EGH)

def scale_decomposed_elutions(x, y, mapped_y, mapped_max_y, opt_recs, model, logger=None, debug_info=None):
    """
        mapped_y may be mapped_uv_vector or corrected xray_vector.
        opt_recs may be updated in abnormal cases where it results in a too small amplitude.
    """
    assert len(opt_recs) > 0

    if debug_info is not None:
        from OptRecsUtils import debug_plot_opt_recs
        ecurve, title = debug_info
        debug_plot_opt_recs(ecurve, opt_recs, title=title + ": entry")

    tmp_recs = copy.deepcopy(opt_recs)

    copied_y = copy.deepcopy(mapped_y)

    init_height_params = []
    for j, rec in enumerate(opt_recs):
        k   = rec[0]
        if k < 0:
            # do minor elements later
            init_height_params.append(None)
            continue

        fnc = rec[1]
        top_x = rec[3].top_x
        scale = mapped_y[top_x] / y[top_x]
        height = fnc( top_x ) * scale
        copied_y -= fnc(x) * scale
        init_height_params.append( height )

    for j, rec in enumerate(opt_recs):
        k   = rec[0]
        if k >= 0:
            continue

        fnc = rec[1]
        top_x = rec[3].top_x

        # get minor elements init params from the residual
        scale = copied_y[top_x] / y[top_x]
        height = fnc( top_x ) * scale
        init_height_params[j] = height

    def obj_func(height_params, debug=False):
        if debug:
            print( 'init_height_params=', init_height_params )
            fig = plt.figure()
            ax  = fig.gca()
            ax.set_title( "decompose_uv_elution debug: model=" + model.get_name() )
            ax.plot( x[[0, -1]], [mapped_max_y, mapped_max_y], ':', color='red', label='mapped_max_y' )
            ax.plot( x, mapped_y, color='blue', label='data' )

        try:
            resid_y = copy.deepcopy(mapped_y)
            for k, h in enumerate(height_params):
                rec = tmp_recs[k]
                fnc = rec[1]
                fnc.update_param(0, h)
                fy = fnc(x)
                if debug:
                    ax.plot(x, fy, label='%s' % str(k))
                resid_y -= fy

            if debug:
                ax.plot( x, resid_y, ':', label='residuals' )
                ax.legend()
                fig.tight_layout()
                plt.show()

            error = np.sum( resid_y**2 )
        except:
            error = np.inf

        return error

    result = minimize( obj_func, np.array(init_height_params) )

    if False:
        obj_func(result.x, debug=True)

    ret_recs = []
    to_remove = []
    for k, h in enumerate(result.x):
        rec = tmp_recs[k]
        fnc = rec.evaluator
        fnc.update_param(0, h)
        if rec.peak.sign > 0:
            max_h = np.max(fnc(x))
        else:
            max_h = np.min(fnc(x))
        max_h_ratio = max_h/mapped_max_y
        # print([k], 'max_h_ratio=', max_h_ratio)
        if abs(max_h_ratio) < REMOVE_H_RATIO_LIMIT:
            """
            such as the 3rd element of 20180204/Close01
            """
            to_remove.append([k, max_h_ratio])
            continue

        ret_recs.append(rec)

    for k, ampl_ratio in reversed(to_remove):
        opt_recs.pop(k)
        if logger is not None:
            logger.warning('%s elution element removed due to too small amplitude ratio %.3g' % (ordinal_str(k+1), ampl_ratio) )

    if False:
        obj_func(result.x, debug=True)

    if debug_info is not None:
        from OptRecsUtils import debug_plot_opt_recs
        ecurve, title = debug_info
        debug_plot_opt_recs(ecurve, opt_recs, title=title + ": result")

    return Struct(opt_recs=ret_recs)



"""
    GuinierTools.RgCurveDbMaker.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import logging

class RgCurveDbMaker:
    def __init__(self, db_folder):
        assert os.path.exists(db_folder)
        self.db_folder = db_folder
        logfile = os.path.join(self.db_folder, "dbmaker.log")
        logging.basicConfig(filename=logfile,
                            level=logging.INFO,
                            format='%(asctime)s %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S')
        self.logger = logging.getLogger(__name__)

    def make_rg_curves(self, restart_count=None):
        from molass_legacy.Batch.DataWalk import test_data_walk
        print("make_rg_curves: ")
        self.restart_count = restart_count

        counters = [0]
        test_data_walk(self.make_call_back, counters=counters)

    def make_call_back(self, in_folder, batch, count, **kwargs):
        from molass_legacy.SerialAnalyzer.DataUtils import cut_upper_folders
        from RgProcess.RgCurve import RgCurve

        if self.restart_count is None:
            pass
        else:
            if count < self.restart_count:
                self.logger.info("make_call_back: skipping %s, %s" % (in_folder, count))
                return True

        print("make_call_back: ", in_folder, count, kwargs)
        sd = batch.load_data(in_folder)
        batch.prepare(sd)
        D_, E_, qv_, xr_curve_ = batch.corrected_sd.get_xr_data_separate_ly()
        rg_curve = RgCurve(qv_, xr_curve_, D_, E_)
        folder = cut_upper_folders(in_folder)
        save_folder = os.path.join(self.db_folder, folder)
        os.makedirs(save_folder, exist_ok=True)
        rg_curve.export(save_folder)
        self.logger.info("make_call_back: %s, %s, ok", in_folder, count)
        return True

if __name__ == '__main__':
    import sys
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.join(this_dir, '..')
    sys.path.append(lib_dir)
    maker = RgCurveDbMaker(r'D:\PyTools\RgCurves')
    maker.make_rg_curves(restart_count=67)


"""
    SerialDataLoader.py

    a) There is only one SerialDataLoader object, which belongs to the GuiMain.
    b) It holds a set of serial data which is currently processed.
    c) The current set of data never changes while the user operates on the same input folder.
    d) For such abnormal data as have been inccured by bubbles,
       fixing of them is done only once just after the initial reading.
    e) Flow change recongnition is performed only once, while it can be modified with the helper info.
    f) Baseline correction depends on the set of optimization parameters.
    g) 
            GuiMain -> Analyzer --(1)-> Mapper
                                <------
                                --(2)-> Executor
                                <------
    f) For each set of optimization parameters, 
    f) The current set is copied every time to be handed to the Anlyzer.
    g) While the 

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""

import os
import copy
import threading
import logging
import time
from SerialDataUtils import get_uv_filename, load_intensity_files, load_uv_array
from SerialData import REQUIRED_SERIAL_NUM_FILES, SerialData
from molass_legacy._MOLASS.SerialSettings import get_setting

DEBUG_LOG       = True

def get_restrict_ranges():
        u_lower = get_setting( 'uv_restrict_lower' )
        u_upper = get_setting( 'uv_restrict_upper' )
        x_lower = get_setting( 'xray_restrict_lower' )
        x_upper = get_setting( 'xray_restrict_upper' )
        return [ (u_lower, u_upper), (x_lower, x_upper) ]

class SerialDataLoader:
    def __init__(self, dialog=None):
        self.logger         = logging.getLogger( __name__ )
        self.dialog = dialog
        self.in_folder      = None
        self.uv_folder      = None
        self.uv_file        = None
        self.load_thread    = None
        self.reset_input_status()
        self.reset_load_status()

    def reset_input_status( self ):
        self.lvector_       = None
        self.uv_array_      = None
        self.uv_file_       = None
        self.col_header     = None
        self.lvector_xp     = None
        self.uv_array_xp    = None
        self.uv_file_xp     = None
        self.is_loadable    = False
        self.is_busy        = False

    def reset_load_status( self ):
        self.is_loadable    = True
        self.ready          = False
        self.initial_data   = None
        self.current_data   = None
        self.is_restricted  = None
        self.is_busy        = True
        self.mtd_elution    = None

    def reset_current_status( self ):
        self.current_data   = None

    def load_from_folders( self, in_folder, uv_folder=None, uv_file=None ):

        if uv_folder is None:
            uv_folder   = in_folder

        assert uv_file is not None or uv_folder is not None

        if uv_file is None:
            uv_file     = get_uv_filename( uv_folder )

        if (    in_folder == self.in_folder
            and uv_folder == self.uv_folder
            and uv_file == self.uv_file
            and self.ready
            ):
            self.logger.info( 'skipped loading for the same input params.' )
            return

        self.in_folder  = in_folder
        self.uv_folder  = uv_folder
        self.uv_file    = uv_file
        self.reset_load_status()

        self.load_thread = threading.Thread(
                                target=self.load_data,
                                name='LoadDataThread',
                                args=[  in_folder,
                                        uv_folder,
                                        uv_file
                                     ],
                                )
        self.load_thread.start()
        self.start_time = time.time()
        self.logger.info( 'started loading ' + in_folder )

    def load_data( self, in_folder, uv_folder, uv_file ):
        self.logger.info( 'loading by SerialDataLoader' )
        self.load_xray_data( in_folder )
        self.load_uv_data( uv_folder, uv_file )
        self.ready = True
        self.is_busy = False    # added for the Tester to be robust

    def wait_until_ready( self ):
        assert self.load_thread is not None

        while not self.ready:
            if self.dialog is None:
                time.sleep(0.1)
            else:
                self.dialog.update()

        self.load_thread.join()
        load_time = time.time() - self.start_time
        self.logger.info( 'finished loading. xray_array.shape=%s. it took %.3g seconds for loading.', str(self.xray_array.shape), load_time )
        self.is_busy = False

    def load_xray_data_only( self, in_folder ):
        self.in_folder  = in_folder
        self.uv_folder  = None
        self.uv_file    = None
        self.reset_load_status()

        self.load_thread = threading.Thread(
                                target=self.load_xray_data_only_impl,
                                name='LoadDataThread',
                                args=[  in_folder,
                                     ],
                                )
        self.load_thread.start()
        self.start_time = time.time()
        self.logger.info('started loading (xray only) ' + in_folder )

    def load_xray_data_only_impl( self, in_folder ):
        self.load_xray_data( in_folder )
        if get_setting('use_xray_conc'):
            if self.has_enough_num_files():
                self.make_xray_proportinal_uv_data()
                self.logger.info('made xray proportinal uv data')
        elif get_setting('use_mtd_conc'):
            self.make_mtd_simulated_uv_data(len(self.datafiles))
            self.logger.info('made mtd simulated uv data')
        else:
            self.logger.error('unexpected state: no UV data')
        self.ready  = True
        self.is_busy = False    # added for the Tester to be robust

    def load_xray_data( self, in_folder ):
        if not os.path.exists( in_folder ):
            raise Exception( in_folder + ' does not exist!'  )

        data_array, datafiles = load_intensity_files( in_folder, logger=self.logger )
        self.datafiles  = datafiles
        self.xray_array = data_array
        self.excluded_xray_array = None

    def load_uv_data( self, uv_folder, uv_file ):
        data_array, lvector, uv_file_, col_header = load_uv_array( uv_folder, uv_file, column_header=True )
        self.uv_array_  = data_array
        self.lvector_   = lvector
        self.uv_file_   = uv_file_
        self.col_header = col_header

    def get_current_object( self ):
        if self.current_data is None:
            self.wait_until_ready()
            self.get_current_object_impl()

        return self.current_data 

    def get_current_object_impl( self ):
        xray_array = self.xray_array
        lvector, uv_array, uv_file, col_header = self.get_active_uv_data()
        data_info   = [ self.datafiles, xray_array, uv_array, lvector, col_header, self.mtd_elution ]
        self.current_data = SerialData( self.uv_folder, self.in_folder, conc_file=uv_file, data_info=data_info )

    def get_data_object( self ):
        if self.current_data is not None:
            if not self.is_restricted:
                return self.current_data
            else:
                pass
        else:
            pass

        self.get_current_object()
        self.initial_data = self.current_data

        self.is_restricted      = False
        self.restrict_ranges    = None

        return self.current_data

    def get_init_object( self ):
        if self.initial_data is None:
            self.get_data_object()
        return self.initial_data

    def get_active_uv_data( self ):
        if get_setting( 'use_xray_conc' ) == 0 and get_setting( 'use_mtd_conc' ) == 0:
            lvector     = self.lvector_
            uv_array    = self.uv_array_
            uv_file     = self.uv_file_
            col_header  = self.col_header
        else:
            lvector     = self.lvector_xp
            uv_array    = self.uv_array_xp
            uv_file     = self.uv_file_xp
            col_header  = None

        # TODO: verify this setting
        self.lvector    = lvector
        self.uv_array   = uv_array
        self.uv_file    = uv_file
        self.col_header = col_header

        return lvector, uv_array, uv_file, col_header

    def get_restricted_uv_data( self, lower, upper ):
        upper_ = None if upper is None else upper+1
        lower_ = None if lower is None else lower
        slice_  = slice( lower_, upper_ )
        lvector, uv_array, uv_file, col_header = self.get_active_uv_data()
        return uv_array[:, slice_], lvector, col_header[slice_]

    def get_restricted_xray_data( self, lower, upper ):
        print( 'get_restricted_xray_data' )
        upper_ = None if upper is None else upper+1
        lower_ = None if lower is None else lower
        slice_  = slice( lower_, upper_ )

        xray_array = self.xray_array
        if self.excluded_xray_array is not None:
            # print( 'getting restricted data from excluded_xray_array' )
            xray_array = self.excluded_xray_array

        return self.datafiles[slice_], xray_array[slice_, :, :]

    def load_xray_data_in_another_thread( self, in_folder ):
        if not os.path.exists( in_folder ):
            raise Exception( in_folder + ' does not exist!'  )

        self.load_xray_thread = threading.Thread(
                                target=self.load_xray_data,
                                name='LoadDataThread',
                                args=[  in_folder,
                                     ],
                                )
        self.load_xray_thread.start()
        self.start_time = time.time()
        print( 'started xray loading.' )

    def wait_for_xray_loading( self ):
        assert self.load_xray_thread is not None

        self.load_xray_thread.join()
        load_time = time.time() - self.start_time
        print( 'finished xray loading. it took %.3g seconds for loading.', load_time)

    def memorize_exclusion( self, serial_data ):
        # print( 'memorize_exclusion' )
        self.excluded_xray_array = serial_data.intensity_array

    def has_enough_num_files( self ):
        return self.xray_array.shape[0] >= REQUIRED_SERIAL_NUM_FILES

    def make_xray_proportinal_uv_data(self, ivector=None):
        from molass_legacy.UV.XrayProportional import make_xray_proportinal_uv_data_impl

        self.logger.info('making xray_proportinal_uv_data')

        lvector_xp, uv_array_xp, uv_file_xp = make_xray_proportinal_uv_data_impl(self.xray_array, ivector=ivector)

        self.logger.info('made uv_array_xp shape=%s from xray_array shape=%s', str(uv_array_xp.shape), str(self.xray_array.shape))

        self.lvector_xp = lvector_xp
        self.uv_array_xp = uv_array_xp
        self.uv_file_xp = uv_file_xp

    def make_mtd_simulated_uv_data( self, size ):
        self.logger.info('making mtd_simulated_uv_data')
        from Microfluidics.MicrofluidicElution import MicrofluidicElution
        mtd_file_path = get_setting( 'mtd_file_path' )
        self.mtd_elution = MicrofluidicElution(mtd_file_path, size)
        x, y = self.mtd_elution.get_elution_data(size)
        self.make_xray_proportinal_uv_data(ivector=y)



"""

    Rank.SrrTutor.py

    Copyright (c) 2022, SAXS Team, KEK-PF

"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from matplotlib.gridspec import GridSpec
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Batch.StandardProcedure import StandardProcedure
from molass_legacy.DataStructure.MatrixData import simple_plot_3d
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def rankview_plot(fig, in_folder, sd, M, vec, data_type, xr_type,
                    pre_recog=None, trim=False, i_only=False, j_only=False,
                    data_label="",
                    view_kw=None, return_axes=False):
    if trim:
        if pre_recog is None:
            from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
            pre_recog = PreliminaryRecognition(sd)
        if xr_type:
            islice, jslice = pre_recog.get_xr_slices()
        else:
            islice, jslice = pre_recog.get_uv_slices()
        if i_only:
            jslice = slice(None, None)
        if j_only:
            islice = slice(None, None)
        D = M[islice,jslice]
        vec = vec[islice]
        trim_label = " trim:" + str([(s.start, s.stop) for s in [islice, jslice]]).replace("None", "_")
        trim_label = trim_label.replace("[", "").replace("]", "")
    else:
        D = M
        trim_label = ""

    N = np.linalg.norm(D)

    U, s, VT = np.linalg.svd(D)
    reconst = []
    residuals = []
    residuals_norms = []
    for rank in range(1,5):
        D_ = U[:,0:rank] @ np.diag(s[0:rank]) @ VT[0:rank,:]
        reconst.append(D_)
        R = D - D_
        residuals.append(R)
        residuals_norms.append(np.linalg.norm(R))

    gs = GridSpec(5,12)
    in_folder = get_in_folder(in_folder)
    fig.suptitle("SVD Reconstruction with various Ranks of %s (%s%s)%s" % (in_folder, data_label, data_type, trim_label), fontsize=24)
    ax0 = fig.add_subplot(gs[0:4,0:4], projection="3d")
    ax0.set_title("Whole Data", fontsize=20)
    ax0.set_axis_off()

    def my_plot_3d(ax, D, x, view_kw=None):
        simple_plot_3d(ax, D, x=x)
        ax.margins(0)
        if view_kw is not None:
            ax.view_init(**view_kw)

    my_plot_3d(ax0, D, vec, view_kw=view_kw)

    axes = []
    for i in range(2):
        i2 = i*2
        axes_row = []
        if i == 0:
            z_limits = []
        else:
            z_limits = np.array(z_limits)
            zmin = np.min(z_limits[:,0])
            zmax = np.max(z_limits[:,1])
        for j in range(4):
            j2 = 4 + j*2
            ax = fig.add_subplot(gs[i2:i2+2,j2:j2+2], projection="3d")
            ax.set_axis_off()
            if i == 0:
                ax.set_title("Rank %d" % (j+1), fontsize=20)
                my_plot_3d(ax, reconst[j], vec, view_kw=view_kw)
                z_limits.append(ax.get_zlim())
            else:
                my_plot_3d(ax, residuals[j], vec, view_kw=view_kw)
                ax.set_zlim(zmin, zmax)
            axes_row.append(ax)
        axes.append(axes_row)
    axes = np.array(axes)

    ax1 = fig.add_subplot(gs[4,0:4])
    ax1.set_axis_off()

    srr1 = residuals_norms[0]/N
    srr2 = residuals_norms[1]/N
    srr21 = srr2/srr1
    ax1.text(0.5, 0.7, r"$ SRR(1) = %.2g, \;\; SRR(2) = %.2g, \;\; SRR(2:1) = \frac{SRR(2)}{SRR(1)} = %.2g$" % (srr1, srr2, srr21), ha="center", fontsize=14)
    ax1.text(1.0, 0.4, "Norms of residual matrices are plotted in the right figures.", ha="right", fontsize=14)
    ax1.text(1.0, 0.2, r"Note that $ \| A + B \| \leq \| A \| + \| B \| $.", ha="right", fontsize=14)

    b_axes = []
    b_twin_axes = []
    y_limits = []
    for j in range(4):
        j2 = 4 + j*2
        ax = fig.add_subplot(gs[4,j2:j2+2])
        ax.set_axis_off()
        ax.bar(0, residuals_norms[j], width=0.5, alpha=0.5, label="residual norm (relative)")
        ax.legend()
        y_limits.append(ax.get_ylim())
        b_axes.append(ax)

        axt = ax.twinx()
        axt.set_axis_off()
        axt.bar(0, residuals_norms[j], width=0.5, alpha=0.5, color="C1", label="residual norm (proportion)")
        axt.legend(bbox_to_anchor=(1, 0.8), loc="upper right")
        b_twin_axes.append(axt)

    y_limits = np.array(y_limits)
    ymin = np.min(y_limits[:,0])
    ymax = np.max(y_limits[:,1])
    Nymax = N*ymax/residuals_norms[0]
    for j in range(4):
        ax = b_axes[j]
        ax.set_ylim(ymin, ymax)
        axt = b_twin_axes[j]
        axt.set_ylim(ymin, Nymax)

    fig.tight_layout()
    fig.subplots_adjust(top=0.88, left=0.01)

    if return_axes:
        return ax0, axes, ax1, b_axes, b_twin_axes

def srr_turtorial(in_folder, **kwargs):
    set_setting("test_pattern", 0)
    sp = StandardProcedure()
    sp.load_old_way(in_folder)
    correct = kwargs.pop("correct", False)
    trim = kwargs.pop("trim", False)
    if correct:
        sd = sp.get_corrected_sd(proxy=False)
    else:
        sd = sp.get_sd(whole=True)

    data_type = kwargs.pop("data_type", "xr")
    xr_type = data_type.lower().find("x") >= 0
    if xr_type:
        M, E, vec, ecurve = sd.get_xr_data_separate_ly()
    else:
        M, _, vec, ecurve = sd.get_uv_data_separate_ly()

    fig = plt.figure(figsize=(20,9))
    rankview_plot(fig, in_folder, sd, M, vec, data_type, xr_type, **kwargs)
    plt.show()



"""

    LPM.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF

"""
import numpy as np
import logging
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from scipy.interpolate import LSQUnivariateSpline
from molass_legacy.SerialAnalyzer.BasePercentileOffset import base_percentile_offset
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.DataStructure.MatrixData import simple_plot_3d
from molass_legacy.SerialAnalyzer.ElutionBaseCurve import ElutionBaseCurve
from molass_legacy.Baseline.LambertBeer import BasePlane
from molass_legacy.Baseline.BaseScattering import get_baseplane

def get_corrected(y, x=None):
    if x is None:
        x = np.arange(len(y))
    sbl = ScatteringBaseline(y, x=x)
    A, B = sbl.solve()
    return y - (A*x+B)

def compute_base_surface(shape, line0, line1):
    n = shape[1]
    w = np.arange(n)/(n-1)
    return (1-w)*line0[:,np.newaxis] + w*line1[:,np.newaxis]

def debug_plot(data, base_surface=None):
    import molass_legacy.KekLib.DebugPlot as plt
    plt.push()
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    simple_plot_3d(ax, data)
    if base_surface is not None:
        simple_plot_3d(ax, base_surface, alpha=0.5)
    fig.tight_layout()
    plt.show()
    plt.pop()    


class BestPercentile:
    def __init__(self, ecurve):
        self.x = ecurve.x
        knots   = np.linspace( 0, len(self.x), len(self.x)//10 )
        self.iknots  = knots[1:-1]
        self.size_sigma = ecurve.compute_size_sigma()

    def get(self, row):
        spline  = LSQUnivariateSpline(self.x, row, self.iknots)
        noisiness = self.compute_noisiness(spline, row)
        percentile =  base_percentile_offset(noisiness, size_sigma=self.size_sigma)
        return percentile

    def compute_noisiness(self, spline, Z):
        pp  = np.percentile( Z, [95, 5] )
        height  = pp[0] - pp[1]
        noisiness = np.std( Z - spline(self.x) ) / height
        return noisiness

class LPM_3d:
    def __init__(self, data, ecurve_y=None, smoothing=True, integral=False, for_all_q=True, e_index=None, end_slices=None, debug=False):
        # debug_plot(data)
        self.logger = logging.getLogger(__name__)

        if integral:
            from molass_legacy.Baseline.Baseline import integrative_curve
            smoothing = False
            self.logger.info("smoothing will not be applied in LPM, which is yet to be supported with integral.")

        if ecurve_y is None:
            self.logger.info("applying LPM correction with fixed percentile.")
        else:
            self.logger.info("applying LPM correction varying percentile.")
            ebase_curve = ElutionBaseCurve(ecurve_y)
            bp = BestPercentile(ebase_curve)

        if for_all_q:
            if smoothing:
                base_points_list = []
                xlim = data.shape[1] - 1
                end_points = [0, xlim]
                points = None

                for k, row in enumerate(data):
                    try:
                        sbl = ScatteringBaseline(row, suppress_warning=True)
                        if ecurve_y is None:
                            A, B = sbl.solve()
                        else:
                            percentile = bp.get(row)
                            A, B = sbl.solve(p_final=percentile)
                        points = [A*x+B for x in end_points]
                    except Exception as exc:
                        assert points is not None
                        self.logger.warning("previous points [%g, %g] are used as %d-th end_points due to '%s'", *points, k, str(exc))
                    base_points_list.append(points)

                base_points = np.array(base_points_list)
                line0 = smooth(base_points[:,0])
                line1 = smooth(base_points[:,1])
                base_surface = compute_base_surface(data.shape, line0, line1)
                corrected_data = data - base_surface
            else:
                x = np.arange(data.shape[1])
                corrected_data = data.copy()
                for k, row in enumerate(corrected_data):
                    try:
                        sbl = ScatteringBaseline(row, suppress_warning=True)
                        if ecurve_y is None:
                            A, B = sbl.solve()
                        else:
                            percentile = bp.get(row)
                            A, B = sbl.solve(p_final=percentile)
                    except Exception as exc:
                        assert points is not None
                        self.logger.warning("previous params (%g, %g) are used as %d-th end_points due to '%s'", A, B, k, str(exc))

                    baseline = A*x + B
                    if integral:
                        baseline = integrative_curve(row, baseline)
                    row -= baseline
        else:
            assert e_index is not None

            print('e_index=', e_index)
            x = np.arange(data.shape[1])
            y = ecurve_y
            # y = data[e_index,:]
            sbl = ScatteringBaseline(y, suppress_warning=True)
            if ecurve_y is None:
                A, B = sbl.solve()
            else:
                percentile = bp.get(y)
                A, B = sbl.solve(p_final=percentile)

            baseline = A*x + B
            if integral:
                baseline = integrative_curve(y, baseline)

            assert end_slices is not None
            self.logger.info("making integral BP with %s" % str(end_slices))
            dy = np.average(data[:,end_slices[1]], axis=1) - np.average(data[:,end_slices[0]], axis=1)
            BP = get_baseplane(baseline, end_slices, dy)

            if debug:
                import molass_legacy.KekLib.DebugPlot as plt
                plt.push()
                fig = plt.figure()
                ax = fig.add_subplot(projection='3d')

                xm = np.ones(len(x))*e_index

                if True:
                    simple_plot_3d(ax, BP)
                else:
                    simple_plot_3d(ax, data)
                    ax.plot(xm, x, y, color='orange')

                ax.plot(xm, x, baseline, color='red')
                x_ = np.arange(data.shape[0])
                y_ = np.ones(data.shape[0])*(len(y)-1)
                ax.plot(x_, y_, dy, color='green')

                fig.tight_layout()
                plt.show()
                plt.pop()
            corrected_data = data - BP

        self.data = corrected_data

    def adjust_with_mf(self, index, ecurve):
        bp = BasePlane(self.data, index, ecurve, denoise=False)
        bp.solve()
        print('bp.params=', bp.params)
        BP = bp.get_baseplane()
        self.data -= BP
        return BP

def get_corrected_matrix(D):
    xray_baseline_type = get_setting('xray_baseline_type')
    if xray_baseline_type == 0:
        retD = D.copy()
    else:
        integral = xray_baseline_type == 5
        lpm = LPM_3d(D, smoothing=True, integral=integral, for_all_q=True)
        retD = lpm.data
    return retD



# coding: utf-8
"""

    OurTtk.py

"""
import tkinter as tk
import tkinter.ttk as ttk

"""
    How to make ttk.Scale behave more like tk.Scale?
    https://stackoverflow.com/questions/47200625/how-to-make-ttk-scale-behave-more-like-tk-scale

    use ttkwidgets.TickScale instead
"""
class TtkScale(ttk.Frame):
    def __init__(self, master=None, **kwargs):
        ttk.Frame.__init__(self, master)
        self.columnconfigure(0, weight=1)
        self.showvalue = kwargs.pop('showvalue', True)
        self.tickinterval = kwargs.pop('tickinterval', 0)
        self.digits = kwargs.pop('digits', '0')
        self.slim = kwargs.pop('slim', False)
        self.showvalue_y = kwargs.pop('showvalue_y', '0')
        self.variable = kwargs.pop('variable', None)
        if self.variable is None:
            self.variable = tk.DoubleVar()

        if 'command' in kwargs:
            # add self.display_value to the command
            fct = kwargs['command']

            def cmd(value):
                fct(value)
                self.display_value(value)

            kwargs['command'] = cmd
        else:
            kwargs['command'] = self.display_value

        self.scale = ttk.Scale(self, **kwargs)
        self.scale.config(variable=self.variable)

        # get slider length
        style = ttk.Style(self)
        style_name = kwargs.get('style', '%s.TScale' % (str(self.scale.cget('orient')).capitalize()))
        self.sliderlength = style.lookup(style_name, 'sliderlength', default=30)

        self.extent = kwargs['to'] - kwargs['from_']
        self.start = kwargs['from_']
        # showvalue
        if self.showvalue:
            if not self.slim:
                ttk.Label(self, text=' ').grid(row=0)
            self.label = ttk.Label(self, text='0')
            self.label.place(in_=self.scale, bordermode='outside', x=0, y=self.showvalue_y, anchor='s')
            self.display_value(self.scale.get())

        self.scale.grid(row=0 if self.slim else 1, sticky='ew')

        # ticks
        if self.tickinterval:
            ttk.Label(self, text=' ').grid(row=2)
            self.ticks = []
            self.ticklabels = []
            nb_interv = round(self.extent/self.tickinterval)
            formatter = '{:.' + str(self.digits) + 'f}'
            for i in range(nb_interv + 1):
                tick = kwargs['from_'] + i * self.tickinterval
                self.ticks.append(tick)
                self.ticklabels.append(ttk.Label(self, text=formatter.format(tick)))
                self.ticklabels[i].place(in_=self.scale, bordermode='outside', x=0, rely=1, anchor='n')
            self.place_ticks()

        self.scale.bind('<Configure>', self.on_configure)

    def convert_to_pixels(self, value):
        return ((value - self.start)/ self.extent) * (self.scale.winfo_width()- self.sliderlength) + self.sliderlength / 2

    def display_value(self, value):
        # position (in pixel) of the center of the slider
        x = self.convert_to_pixels(float(value))
        # pay attention to the borders
        half_width = self.label.winfo_width() / 2
        if x + half_width > self.scale.winfo_width():
            x = self.scale.winfo_width() - half_width
        elif x - half_width < 0:
            x = half_width
        self.label.place_configure(x=x)
        formatter = '{:.' + str(self.digits) + 'f}'
        self.label.configure(text=formatter.format(float(value)))

    def place_ticks(self):
        if not self.tickinterval:
            return
        # first tick 
        tick = self.ticks[0]
        label = self.ticklabels[0]
        x = self.convert_to_pixels(tick)
        half_width = label.winfo_width() / 2
        if x - half_width < 0:
            x = half_width
        label.place_configure(x=x)
        # ticks in the middle
        for tick, label in zip(self.ticks[1:-1], self.ticklabels[1:-1]):
            x = self.convert_to_pixels(tick)
            label.place_configure(x=x)
        # last tick
        tick = self.ticks[-1]
        label = self.ticklabels[-1]
        x = self.convert_to_pixels(tick)
        half_width = label.winfo_width() / 2
        if x + half_width > self.scale.winfo_width():
            x = self.scale.winfo_width() - half_width
        label.place_configure(x=x)

    def on_configure(self, event):
        """Redisplay the ticks and the label so that they adapt to the new size of the scale."""
        self.display_value(self.scale.get())
        self.place_ticks()

    def set(self, value):
        self.variable.set(value)

    def get(self):
        return self.variable.get()



"""
    Optimizer.XrStateFigure.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DataUtils import cut_upper_folders
from .FvScoreConverter import convert_score

DATA_NAME_DICT = {
    "20220210": "C1015s",
    "20220517": "C1015F",
    "20240730": "5913s001",
}

def show_this_figure_impl(self, best=True):
    optimizer = self.fullopt
    modelname = optimizer.get_model_name()
    if best:
        params = self.get_best_params()
    else:
        params = self.get_current_params()
    fv = optimizer.objective_func(params)
    sv = convert_score(fv)
    in_folder = get_setting('in_folder')
    folder = cut_upper_folders(in_folder)
    dataname = DATA_NAME_DICT[folder]
    with plt.Dp(button_spec=["Close"]):
        fig, ax = plt.subplots(figsize=(6,4.5))
        ax.set_title("Decomposition of %s with %s, SV=%.1f" % (dataname, modelname, sv), fontsize=16, y=1.01)
        axt = ax.twinx()
        axt.grid(False)
        axes = (None, ax, None, axt)
        optimizer.objective_func(params, plot=True, axis_info=(fig, axes))
        fig.tight_layout()
        plt.show()